from functools import lru_cache
from typing import Iterator, Optional

from hiku.context import ExecutionContext
from hiku.endpoint.graphql import _run_validation
from hiku.extensions.base_extension import Extension


class QueryValidationCache(Extension):
    """Sets up lru cache for the query Node validation.

    :param int maxsize: Maximum size of the cache
    """

    def __init__(self, maxsize: Optional[int] = None):
        self.cached_validator = lru_cache(maxsize=maxsize)(_run_validation)

    def on_validate(
        self, execution_context: ExecutionContext
    ) -> Iterator[None]:
        execution_context.errors = self.cached_validator(
            execution_context.graph,
            execution_context.query,
            execution_context.validators,
        )

        yield



import hashlib
from typing import Iterator

from sentry_sdk import configure_scope, start_span

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension


class SentryTracing(Extension):
    def get_resource_name(self, execution_context: ExecutionContext) -> str:
        assert execution_context.query

        query_hash = self._hash_query(execution_context.query_src)

        if execution_context.operation_name:
            return f"{execution_context.operation_name}:{query_hash}"

        return query_hash

    def _hash_query(self, query: str) -> str:
        return hashlib.md5(query.encode("utf-8")).hexdigest()

    def on_dispatch(
        self, execution_context: ExecutionContext
    ) -> Iterator[None]:
        self._operation_name = execution_context.operation_name
        name = self._operation_name or "Anonymous Query"

        with configure_scope() as scope:
            if scope.span:
                self.gql_span = scope.span.start_child(
                    op="gql", description=name
                )
            else:
                self.gql_span = start_span(op="gql")

        operation_type = execution_context.operation_type_name

        self.gql_span.set_tag("graphql.operation_type", operation_type)
        self.gql_span.set_tag(
            "graphql.resource_name", self.get_resource_name(execution_context)
        )
        self.gql_span.set_data("graphql.query", execution_context.query_src)

        yield

        self.gql_span.finish()

    def on_validate(
        self, execution_context: ExecutionContext
    ) -> Iterator[None]:
        self.validation_span = self.gql_span.start_child(
            op="validation", description="Validation"
        )

        yield

        self.validation_span.finish()

    def on_parse(self, execution_context: ExecutionContext) -> Iterator[None]:
        self.parsing_span = self.gql_span.start_child(
            op="parsing", description="Parsing"
        )

        yield

        self.parsing_span.finish()



from abc import abstractmethod
from asyncio import gather
from inspect import isawaitable
from typing import (
    List,
    Dict,
    Any,
    Optional,
    Sequence,
    Tuple,
    cast,
    overload,
    Union,
    Type,
)

from hiku.context import (
    ExecutionContext,
    ExecutionContextFinal,
    create_execution_context,
)
from hiku.extensions.base_extension import Extension, ExtensionsManager
from hiku.extensions.base_validator import QueryValidator
from hiku.federation.utils import get_entity_types
from hiku.federation.version import DEFAULT_FEDERATION_VERSION

from hiku.federation.sdl import print_sdl

from hiku.federation.graph import Graph
from hiku.federation.engine import Engine
from hiku.federation.introspection import (
    BaseFederatedGraphQLIntrospection,
    FederatedGraphQLIntrospection,
    AsyncFederatedGraphQLIntrospection,
)

from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.federation.denormalize import DenormalizeEntityGraphQL
from hiku.endpoint.graphql import (
    BaseGraphQLEndpoint,
    BatchedRequest,
    BatchedResponse,
    GraphQLError,
    GraphQLRequest,
    GraphQLResponse,
    SingleOrBatchedRequest,
    SingleOrBatchedResponse,
    _run_validation as _run_validation_base,
)
from hiku.graph import (
    GraphTransformer,
    apply,
    Union as GraphUnion,
    Graph as _Graph,
)
from hiku.query import Node
from hiku.result import Proxy


GraphT = Union[Graph, _Graph]


def _run_validation(
    graph: GraphT,
    query: Node,
    validators: Optional[Tuple[QueryValidator, ...]] = None,
) -> List[str]:
    if "_entities" in query.fields_map or "_service" in query.fields_map:
        return []

    return _run_validation_base(graph, query, validators)


def denormalize_entities(
    graph: Graph,
    query: Node,
    result: Proxy,
) -> List[Dict[str, Any]]:
    representations = query.fields_map["_entities"].options["representations"]

    if not representations:
        return []

    typename = representations[0]["__typename"]
    data = DenormalizeEntityGraphQL(graph, result, typename).process(query)

    return data["_entities"]


class FederationV1EntityTransformer(GraphTransformer):
    def visit_graph(self, obj: Graph) -> Graph:  # type: ignore
        entities = get_entity_types(obj.nodes, federation_version=1)
        unions = []
        for u in obj.unions:
            if u.name == "_Entity" and entities:
                unions.append(GraphUnion("_Entity", entities))
            else:
                unions.append(u)

        return Graph(
            [self.visit(node) for node in obj.items],
            obj.data_types,
            obj.directives,
            unions,
            obj.interfaces,
            obj.enums,
            obj.scalars,
        )


class BaseFederatedGraphEndpoint(BaseGraphQLEndpoint):
    query_graph: Graph
    mutation_graph: Optional[Graph]
    engine: Engine
    batching: bool
    validation: bool
    introspection: bool
    federation_version: int

    @property
    @abstractmethod
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        pass

    def __init__(
        self,
        engine: Engine,
        query_graph: Graph,
        mutation_graph: Optional[Graph] = None,
        batching: bool = False,
        validation: bool = True,
        introspection: bool = True,
        extensions: Optional[
            Sequence[Union[Extension, Type[Extension]]]
        ] = None,
        federation_version: int = DEFAULT_FEDERATION_VERSION,
    ):
        self.engine = engine
        self.federation_version = federation_version
        self.introspection = introspection
        self.batching = batching
        self.validation = validation
        self.extensions = extensions or []

        execution_context = create_execution_context()
        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )
        with extensions_manager.graph():
            transformers: List[GraphTransformer] = list(
                execution_context.transformers
            )

        if federation_version == 1:
            transformers.append(FederationV1EntityTransformer())

        if self.introspection:
            transformers.append(
                self.introspection_cls(query_graph, mutation_graph)
            )

        self.query_graph = apply(query_graph, transformers)
        if mutation_graph is not None:
            self.mutation_graph = apply(mutation_graph, transformers)
        else:
            self.mutation_graph = None


class BaseSyncFederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    @abstractmethod
    def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> Dict:
        pass

    def dispatch(self, data: Dict, context: Optional[Dict] = None) -> Dict:
        execution_context = create_execution_context(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
            query_graph=self.query_graph,
            mutation_graph=self.mutation_graph,
        )

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.dispatch():
                self._init_execution_context(
                    execution_context, extensions_manager
                )
                result = self.execute(execution_context, extensions_manager)
                return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class BaseAsyncFederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    @abstractmethod
    async def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> Dict:
        pass

    async def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse:
        execution_context = create_execution_context(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
            query_graph=self.query_graph,
            mutation_graph=self.mutation_graph,
        )

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.dispatch():
                self._init_execution_context(
                    execution_context, extensions_manager
                )
                result = await self.execute(
                    execution_context, extensions_manager
                )
                return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class FederatedGraphQLEndpoint(BaseSyncFederatedGraphQLEndpoint):
    """Can execute either regular or federated queries.
    Handles following fields of federated query:
        - _service
        - _entities
    """

    @property
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        return FederatedGraphQLIntrospection

    def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> Dict:
        execution_context = cast(ExecutionContextFinal, execution_context)

        with extensions_manager.validation():
            if self.validation and execution_context.errors is None:
                execution_context.errors = _run_validation(
                    execution_context.graph,
                    execution_context.query,
                    execution_context.validators,
                )

        if execution_context.errors:
            raise GraphQLError(errors=execution_context.errors)

        with extensions_manager.execution():
            if "_service" in execution_context.query.fields_map:
                execution_context.context["__sdl__"] = print_sdl(
                    self.query_graph,
                    self.mutation_graph,
                    self.federation_version,
                )

            result = self.engine.execute_context(execution_context)
            assert isinstance(result, Proxy)
            execution_context.result = result

        return DenormalizeGraphQL(
            execution_context.graph,
            result,
            execution_context.operation_type_name,
        ).process(execution_context.query)

    def dispatch(self, data: Dict, context: Optional[Dict] = None) -> Dict:
        execution_context = create_execution_context(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
            query_graph=self.query_graph,
            mutation_graph=self.mutation_graph,
        )

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.dispatch():
                self._init_execution_context(
                    execution_context, extensions_manager
                )
                result = self.execute(execution_context, extensions_manager)
                return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class AsyncFederatedGraphQLEndpoint(BaseAsyncFederatedGraphQLEndpoint):
    @property
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        return AsyncFederatedGraphQLIntrospection

    async def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> Dict:
        execution_context = cast(ExecutionContextFinal, execution_context)

        with extensions_manager.validation():
            if self.validation and execution_context.errors is None:
                execution_context.errors = _run_validation(
                    execution_context.graph,
                    execution_context.query,
                    execution_context.validators,
                )

        if execution_context.errors:
            raise GraphQLError(errors=execution_context.errors)

        with extensions_manager.execution():
            if "_service" in execution_context.query.fields_map:
                execution_context.context["__sdl__"] = print_sdl(
                    self.query_graph,
                    self.mutation_graph,
                    self.federation_version,
                )

            coro = self.engine.execute_context(execution_context)
            assert isawaitable(coro)
            result = await coro
            execution_context.result = result

        return DenormalizeGraphQL(
            execution_context.graph,
            result,
            execution_context.operation_type_name,
        ).process(execution_context.query)

    @overload
    async def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse:
        ...

    @overload
    async def dispatch(
        self, data: BatchedRequest, context: Optional[Dict] = None
    ) -> BatchedResponse:
        ...

    async def dispatch(
        self, data: SingleOrBatchedRequest, context: Optional[Dict] = None
    ) -> SingleOrBatchedResponse:
        if isinstance(data, list):
            return list(
                await gather(
                    *(super().dispatch(item, context) for item in data)
                )
            )

        return await super().dispatch(data, context)


class AsyncBatchFederatedGraphQLEndpoint(AsyncFederatedGraphQLEndpoint):
    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["batching"] = True
        super().__init__(*args, **kwargs)



import contextvars
import time
from abc import abstractmethod
from functools import partial
from typing import Dict, Optional

from prometheus_client import Summary

from ..graph import GraphTransformer
from ..engine import pass_context, _do_pass_context
from ..sources.graph import CheckedExpr


_METRIC = None


def _get_default_metric():
    global _METRIC
    if _METRIC is None:
        _METRIC = Summary(
            "graph_field_time",
            "Graph field time (seconds)",
            ["graph", "node", "field"],
        )
    return _METRIC


def _func_field_names(func):
    fields_pos = 1 if _do_pass_context(func) else 0

    def wrapper(*args):
        return func([f.name for f in args[fields_pos]], *args)

    return wrapper


def _subquery_field_names(func):
    def wrapper(fields, *args):
        return func([f.name for _, f in fields], fields, *args)

    return wrapper


class GraphMetricsBase(GraphTransformer):
    root_name = "Root"

    def __init__(
        self,
        name,
        *,
        metric=None,
        ctx_var: Optional[contextvars.ContextVar] = None,
    ):
        self._name = name
        self._metric = metric or _get_default_metric()
        self._ctx = ctx_var
        self._node = None
        self._wrappers: Dict = {}

    @abstractmethod
    def field_wrapper(self, observe, func):
        raise NotImplementedError

    @abstractmethod
    def link_wrapper(self, observe, func):
        raise NotImplementedError

    @abstractmethod
    def subquery_wrapper(self, observe, subquery):
        raise NotImplementedError

    def _observe_fields(self, node_name):
        by_field = {}

        def observe(start_time, field_names, ctx):
            duration = time.perf_counter() - start_time

            for name in field_names:
                try:
                    field_metric = by_field[name]
                except KeyError:
                    field_metric = by_field[name] = self._metric.labels(
                        *self.get_labels(self._name, node_name, name, ctx),
                    )
                field_metric.observe(duration)

        return observe

    def get_labels(
        self, graph_name: str, node_name: str, field_name: str, ctx: dict
    ) -> list:
        return [graph_name, node_name, field_name]

    def _wrap_field(self, node_name, func):
        observe = self._observe_fields(node_name)
        wrapper = self.field_wrapper(observe, func)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        wrapper = _func_field_names(wrapper)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        return wrapper

    def _wrap_link(self, node_name, link_name, func):
        observe = self._observe_fields(node_name)
        wrapper = self.link_wrapper(observe, func)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        wrapper = partial(wrapper, link_name)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        return wrapper

    def _wrap_subquery(self, node_name, subquery):
        observe = self._observe_fields(node_name)
        wrapper = self.subquery_wrapper(observe, subquery)
        wrapper = _subquery_field_names(wrapper)
        wrapper.__subquery__ = lambda: wrapper
        return wrapper

    def visit_node(self, obj):
        self._node = obj
        try:
            return super().visit_node(obj)
        finally:
            self._node = None

    def visit_field(self, obj):
        obj = super().visit_field(obj)
        node_name = self.root_name if self._node is None else self._node.name
        if isinstance(obj.func, CheckedExpr):
            func = obj.func.__subquery__
        else:
            func = obj.func

        wrapper = self._wrappers.get(func)
        if wrapper is None:
            if isinstance(obj.func, CheckedExpr):
                wrapper = self._wrappers[func] = self._wrap_subquery(
                    node_name,
                    func,
                )
            else:
                wrapper = self._wrappers[func] = self._wrap_field(
                    node_name,
                    func,
                )

        if isinstance(obj.func, CheckedExpr):
            obj.func = CheckedExpr(
                wrapper,
                obj.func.expr,
                obj.func.reqs,
                obj.func.proc,
            )
        else:
            obj.func = wrapper
        return obj

    def visit_link(self, obj):
        obj = super().visit_link(obj)
        node_name = self.root_name if self._node is None else self._node.name
        obj.func = self._wrap_link(node_name, obj.name, obj.func)
        return obj


class _SubqueryMixin:
    def subquery_wrapper(self, observe, subquery):
        def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result_proc = subquery(*args)

            def proc_wrapper():
                result = result_proc()
                ctx = self._ctx.get() if self._ctx else None
                observe(start_time, field_names, ctx)
                return result

            return proc_wrapper

        return wrapper


class GraphMetrics(_SubqueryMixin, GraphMetricsBase):
    def field_wrapper(self, observe, func):
        def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result = func(*args)
            ctx = self._ctx.get() if self._ctx else None
            observe(start_time, field_names, ctx)
            return result

        return wrapper

    def link_wrapper(self, observe, func):
        def wrapper(link_name, *args):
            start_time = time.perf_counter()
            result = func(*args)
            ctx = self._ctx.get() if self._ctx else None
            observe(start_time, [link_name], ctx)
            return result

        return wrapper


class AsyncGraphMetrics(_SubqueryMixin, GraphMetricsBase):
    def field_wrapper(self, observe, func):
        async def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result = await func(*args)
            ctx = self._ctx.get() if self._ctx else None
            observe(start_time, field_names, ctx)
            return result

        return wrapper

    def link_wrapper(self, observe, func):
        async def wrapper(link_name, *args):
            start_time = time.perf_counter()
            result = await func(*args)
            ctx = self._ctx.get() if self._ctx else None
            observe(start_time, [link_name], ctx)
            return result

        return wrapper



from functools import lru_cache
from typing import Iterator, Optional

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension
from hiku.readers.graphql import read_operation
from hiku.utils import to_immutable_dict


class QueryTransformCache(Extension):
    """Sets up lru cache for the ast to Node transformation.

    :param int maxsize: Maximum size of the cache
    """

    def __init__(self, maxsize: Optional[int] = None):
        self.cached_operation_reader = lru_cache(maxsize=maxsize)(
            read_operation
        )

    def on_operation(
        self, execution_context: ExecutionContext
    ) -> Iterator[None]:
        execution_context.operation = self.cached_operation_reader(
            execution_context.graphql_document,
            to_immutable_dict(execution_context.variables)
            if execution_context.variables
            else None,
            execution_context.request_operation_name,
        )

        yield



"""
    hiku.readers.simple
    ~~~~~~~~~~~~~~~~~~~

    Support for queries encoded using EDN format

"""
import typing as t

from ..edn import loads, Dict, List, Keyword, Tuple
from ..query import Node, Link, Field, merge


def _get_options(value):
    if len(value) < 2:
        raise TypeError("Missing options argument")
    elif len(value) > 2:
        raise TypeError("More arguments than expected")

    keyword_value, options_value = value
    if not isinstance(keyword_value, Keyword):
        raise TypeError(
            "Names should be specified as keywords, not as {!r}".format(
                type(keyword_value)
            )
        )

    if not isinstance(options_value, Dict):
        raise TypeError(
            "Options should be specified as mapping, not as {!r}".format(
                type(options_value)
            )
        )

    non_keyword = set(
        (k, type(k)) for k in options_value.keys() if not isinstance(k, Keyword)
    )
    if non_keyword:
        keys_repr = " ".join("{} {!r}".format(k, t) for k, t in non_keyword)
        raise TypeError(
            "Option names should be specified as keywords: {}".format(keys_repr)
        )

    name = str(keyword_value)
    options = {str(k): v for k, v in options_value.items()}
    return name, options


def _extract(values):
    for value in values:
        if isinstance(value, Tuple):
            name, options = _get_options(value)
            yield Field(name, options)
        elif isinstance(value, Keyword):
            yield Field(str(value))
        elif isinstance(value, Dict):
            for key, val in value.items():
                if isinstance(key, Tuple):
                    name, options = _get_options(key)
                elif isinstance(key, Keyword):
                    name = str(key)
                    options = None
                else:
                    raise TypeError(
                        "Link name defined not as keyword, "
                        "but as {!r}".format(key)
                    )
                yield Link(name, transform(val), options)
        else:
            raise TypeError("Invalid node member: {!r}".format(value))


def transform(value: t.List) -> Node:
    if isinstance(value, List):
        return merge([Node(list(_extract(value)))])
    else:
        raise TypeError(
            "Node should be defined as vector, "
            "{!r} provided instead".format(value)
        )


def read(src):
    """Reads a query, encoded using EDN format

    Example:

    .. code-block:: python

        query = read('[{(:characters {:limit 100}) [:name]}]')
        result = engine.execute(graph, query)

    :param str src: EDN-encoded data structure
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    edn_ast = loads(src)
    return transform(edn_ast)



import abc
import contextlib
import hashlib

from collections import (
    defaultdict,
    deque,
)
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Union,
    Deque,
    Iterator,
    Optional,
    Callable,
)

from prometheus_client import Counter

from hiku.compat import Protocol
from hiku.result import Index
from hiku.graph import (
    Many,
    Graph,
    Node,
    Field,
)
from hiku.query import (
    QueryVisitor,
    Field as QueryField,
    Link as QueryLink,
    Node as QueryNode,
)

if TYPE_CHECKING:
    from hiku.engine import Context


RESULT_CACHE_HITS = Counter(
    name="hiku_result_cache_hits",
    documentation="Resolver result cache hits",
    labelnames=["graph", "query_name", "node", "field"],
)
RESULT_CACHE_MISSES = Counter(
    name="hiku_result_cache_misses",
    documentation="Resolver result cache misses",
    labelnames=["graph", "query_name", "node", "field"],
)

CACHE_VERSION = "1"


class Hasher(Protocol):
    def update(self, data: bytes) -> None:
        ...


CacheKeyFn = Callable[["Context", Hasher], None]


class BaseCache(abc.ABC):
    @abc.abstractmethod
    def get_many(self, keys: List[str]) -> Dict[str, Any]:
        """Result must contain only keys which were cached"""
        raise NotImplementedError()

    @abc.abstractmethod
    def set_many(self, items: Dict[str, Any], ttl: int) -> None:
        raise NotImplementedError()


@dataclass
class CacheMetrics:
    name: str
    hits_counter: Counter = RESULT_CACHE_HITS
    misses_counter: Counter = RESULT_CACHE_MISSES


@dataclass
class CacheSettings:
    cache: BaseCache
    cache_key: Optional[CacheKeyFn] = None
    metrics: Optional[CacheMetrics] = None


class CacheInfo:
    __slots__ = ("cache", "cache_key", "metrics", "query_name")

    def __init__(
        self, cache_settings: CacheSettings, query_name: Optional[str] = None
    ):
        self.cache = cache_settings.cache
        self.cache_key = cache_settings.cache_key
        self.metrics = cache_settings.metrics
        self.query_name = query_name or "unknown"

    def _track(self, node: str, field: str, hits: int, misses: int) -> None:
        if not self.metrics:
            return

        self.metrics.hits_counter.labels(
            self.metrics.name, self.query_name, node, field
        ).inc(hits)
        self.metrics.misses_counter.labels(
            self.metrics.name, self.query_name, node, field
        ).inc(misses)

    def query_hash(
        self, ctx: "Context", query_link: QueryLink, req: Any
    ) -> str:
        hasher = hashlib.sha1()
        get_query_hash(hasher, query_link, req)
        if self.cache_key:
            self.cache_key(ctx, hasher)
        return hasher.hexdigest()

    def get_many(
        self, keys: List[str], node: str, field: str
    ) -> Dict[str, Any]:
        data = self.cache.get_many(keys)
        hits = sum(1 for key in keys if key in data)
        misses = len(keys) - hits
        self._track(node, field, hits, misses)
        return data

    def set_many(self, items: Dict[str, Any], ttl: int) -> None:
        self.cache.set_many(items, ttl)


class HashVisitor(QueryVisitor):
    def __init__(self, hasher) -> None:  # type: ignore
        self._hasher = hasher

    def visit_field(self, obj: QueryField) -> None:
        self._hasher.update(obj.index_key.encode())

    def visit_node(self, obj: QueryNode) -> None:
        for field in obj.fields:
            self.visit(field)

        for fr in obj.fragments:
            self.visit(fr)

    def visit_link(self, obj: QueryLink) -> None:
        self._hasher.update(obj.index_key.encode())
        super().visit_link(obj)


class CacheVisitor(QueryVisitor):
    """Visit cached query link to extract all data from index
    that needs to be cached
    """

    def __init__(
        self, cache: CacheInfo, index: Index, graph: Graph, node: Node
    ) -> None:
        self._cache = cache
        self._index = index
        self._graph = graph
        self._node = deque([node])
        self._req: Deque[Any] = deque()
        self._data: Deque[Dict] = deque()
        self._to_cache: Deque[Dict] = deque()
        self._node_idx: Deque[Dict] = deque()

    def visit_field(self, field: QueryField) -> None:
        if field.name == "__typename":
            return

        self._data[-1][field.index_key] = self._node_idx[-1][field.index_key]
        super().visit_field(field)

    def visit_node(self, node: QueryNode) -> None:
        for field in node.fields:
            self.visit(field)

        for fr in node.fragments:
            self.visit(fr)

    def visit_link(self, link: QueryLink) -> None:
        refs = self._node_idx[-1][link.index_key]

        self._data[-1][link.index_key] = refs

        graph_obj = self._node[-1].fields_map[link.name]
        if isinstance(graph_obj, Field):
            # Link as complex field
            return

        node = self._graph.nodes_map[graph_obj.node]
        self._node.append(node)

        @contextlib.contextmanager
        def _visit_ctx(req: Any) -> Iterator:
            self._node_idx.append(self._index[node.name][req])
            self._data.append({})

            yield

            data = self._data.pop()
            self._to_cache[-1][node.name][req] = data
            self._node_idx.pop()

        if graph_obj.type_enum is Many:
            for r in refs:
                with _visit_ctx(r.ident):
                    super().visit_link(link)
        else:
            if refs is None:
                self._node.pop()
                return

            with _visit_ctx(refs.ident):
                super().visit_link(link)

        self._node.pop()

    def process(
        self, link: QueryLink, ids: List, reqs: List, ctx: "Context"
    ) -> Dict:
        to_cache = {}
        for i, req in zip(ids, reqs):
            node = self._node[-1]
            self._node_idx.append(self._index[node.name][i])
            self._data.append({})
            self._to_cache.append(defaultdict(dict))

            self.visit(link)

            self._to_cache[-1][node.name] = self._data.pop()
            key = self._cache.query_hash(ctx, link, req)
            to_cache[key] = dict(self._to_cache.pop())
            self._node_idx.pop()

        return to_cache


def get_query_hash(
    hasher: Hasher, query_link: Union[QueryLink, QueryField], req: Any
) -> None:
    hash_visitor = HashVisitor(hasher)
    hash_visitor.visit(query_link)

    if isinstance(req, list):
        for r in req:
            hasher.update(str(hash(r)).encode("utf-8"))
    else:
        hasher.update(str(hash(req)).encode("utf-8"))
    hasher.update(CACHE_VERSION.encode("utf-8"))



import sys
from typing import NewType, cast

Const = NewType("Const", object)


def const(name: str) -> Const:
    t = type(name, (object,), {})
    t.__module__ = sys._getframe(1).f_globals.get("__name__", "__main__")
    return cast(Const, t)



from typing import (
    Callable,
    Iterable,
    Any,
    List,
)

from . import sqlalchemy as _sa
from ..engine import Context
from ..query import Field

# We are limiting fetch size to reduce CPU usage and avoid event-loop blocking
FETCH_SIZE = 100


class FieldsQuery(_sa.FieldsQuery):
    async def __call__(
        self, ctx: Context, fields_: List[Field], ids: Iterable
    ) -> List:
        if not ids:
            return []

        expr, result_proc = self.select_expr(fields_, ids)

        sa_engine = ctx[self.engine_key]
        async with sa_engine.connect() as connection:
            stream = await connection.stream(expr)
            rows = []
            while True:
                bucket = await stream.fetchmany(FETCH_SIZE)
                if bucket:
                    rows.extend(bucket)
                else:
                    break

        return result_proc(rows)


class LinkQuery(_sa.LinkQuery):
    async def __call__(
        self, result_proc: Callable, ctx: Context, ids: Iterable
    ) -> Any:
        expr = self.select_expr(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.engine_key]
            async with sa_engine.connect() as connection:
                stream = await connection.stream(expr)
                rows = []
                while True:
                    bucket = await stream.fetchmany(FETCH_SIZE)
                    if bucket:
                        rows.extend(bucket)
                    else:
                        break
            pairs = [(r.from_column, r.to_column) for r in rows]
        return result_proc(pairs, ids)



from typing import Iterator, List

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension
from hiku.extensions.base_validator import QueryValidator
from hiku.graph import Graph
from hiku.query import Field, Link, Node


def is_introspection_key(key: str) -> bool:
    """See: https://spec.graphql.org/June2018/#sec-Schema"""
    return key.startswith("__")


class _QueryDepthValidator(QueryValidator):
    max_depth: int

    """
    :param max_depth: maximum allowed query depth
    """

    def __init__(self, max_depth: int):
        self.max_depth = max_depth
        self._current_depth = 0
        self._final_depth = 0

    def validate(self, query: Node, graph: Graph) -> List[str]:
        depth = self.calculate(query)
        if depth > self.max_depth:
            return [
                f"Query depth {depth} exceeds maximum "
                f"allowed depth {self.max_depth}"
            ]

        return []

    def calculate(self, obj: Node) -> int:
        self.visit(obj)
        return self._final_depth

    def visit_field(self, obj: Field) -> None:
        pass  # Do nothing for individual fields.

    def visit_link(self, obj: Link) -> None:
        if not is_introspection_key(obj.name):
            super().visit_link(obj)

    def visit_node(self, obj: Node) -> None:
        self._current_depth += 1
        self._final_depth = max(self._final_depth, self._current_depth)
        super().visit_node(obj)
        self._current_depth -= 1


class QueryDepthValidator(Extension):
    """Use this extension to limit the maximum allowed query depth.

    Example:
        Endpoint(engine, graph, extensions=[
            QueryDepthValidator(max_depth=10),
        ])

    """

    def __init__(self, max_depth: int):
        self._validator = _QueryDepthValidator(max_depth)

    def on_dispatch(
        self, execution_context: ExecutionContext
    ) -> Iterator[None]:
        execution_context.validators = execution_context.validators + tuple(
            [self._validator]
        )

        yield



"""
    hiku.expr.core
    ~~~~~~~~~~~~~~

    Expression building blocks

"""
import typing as t

from functools import wraps
from itertools import chain
from collections import namedtuple

from ..compat import ParamSpec
from ..edn import loads
from ..query import Node as QueryNode, Link, Field, Base as QueryBase
from ..types import (
    Record,
    Callable,
    Any,
    GenericMeta,
)
from ..readers.simple import transform

from .nodes import Symbol, Tuple, List, Keyword, Dict, Node


THIS = "this"

_Func = namedtuple("_Func", "expr, args")


class DotHandler:
    def __init__(self, obj: t.Union[Symbol, Tuple]) -> None:
        self.obj = obj

    def __repr__(self) -> str:
        return repr(self.obj)

    def __getattr__(self, name: str) -> "DotHandler":
        return DotHandler(Tuple([Symbol("get"), self.obj, Symbol(name)]))


# for backward compatibility
_DotHandler = DotHandler


class _S:
    def __getattr__(self, name: str) -> DotHandler:
        return DotHandler(Symbol(name))


#: Helper object to represent symbols in expressions. ``S.foo.bar`` in
#: expressions is equivalent to ``foo.bar`` in the regular Python.
S = _S()


def _to_expr(
    obj: t.Union[DotHandler, _Func, t.List, t.Dict, Node],
    fn_reg: t.Set[t.Callable],
) -> Node:
    if isinstance(obj, DotHandler):
        return obj.obj
    elif isinstance(obj, _Func):
        fn_reg.add(obj.expr)
        values: t.List[Node] = [Symbol(obj.expr.__def_name__)]
        values.extend(_to_expr(arg, fn_reg) for arg in obj.args)
        return Tuple(values)
    elif isinstance(obj, list):
        return List([_to_expr(v, fn_reg) for v in obj])
    elif isinstance(obj, dict):
        values = list(
            chain.from_iterable(
                (Keyword(k), _to_expr(v, fn_reg)) for k, v in obj.items()
            )
        )
        return Dict(values)
    else:
        return obj


def to_expr(
    obj: t.Union[DotHandler, _Func]
) -> t.Tuple[Node, t.Tuple[t.Callable, ...]]:
    functions: t.Set[t.Callable] = set([])
    node = _to_expr(obj, functions)
    return node, tuple(functions)


def _query_to_types(
    obj: QueryBase,
) -> t.Union[t.Type[Any], t.Type[Record]]:
    if isinstance(obj, QueryNode):
        return Record[[(f.name, _query_to_types(f)) for f in obj.fields]]
    elif isinstance(obj, Link):
        return _query_to_types(obj.node)
    elif isinstance(obj, Field):
        return Any
    else:
        raise TypeError(type(obj))


T = t.TypeVar("T")
P = ParamSpec("P")


def define(
    *types: GenericMeta, **kwargs: str
) -> t.Callable[[t.Callable[P, T]], t.Callable[P, _Func]]:
    """Annotates function arguments with types.

    These annotations are used to type-check expressions and to analyze,
    which data is used from provided arguments.

    Example:

    .. code-block:: python

        @define(Record[{'id': Integer, 'name': String}])
        def image_url(image):
            return 'http://example.com/{id}-{name}'.format(id=image['id'],
                                                           name=image['name'])

    Here ``image_url`` function accepts an object as argument, and is using two
    of it's fields: ``id`` field of type ``Integer`` and ``name`` field of type
    ``String``. Hiku will check that this function will be used only with
    objects having at least such two fields.

    This annotation also gives ability for Hiku to build a query for low-level
    graph.
    """

    def decorator(fn: t.Callable[P, T]) -> t.Callable[P, _Func]:
        _name = kwargs.pop("_name", None)
        assert not kwargs, repr(kwargs)

        name = _name or "{}/{}_{}".format(fn.__module__, fn.__name__, id(fn))

        @wraps(fn)
        def expr(*args: P.args, **kw: P.kwargs) -> _Func:
            return _Func(expr, args)

        expr.__def_name__ = name  # type: ignore[attr-defined]
        expr.__def_body__ = fn  # type: ignore[attr-defined]

        if len(types) == 1 and isinstance(types[0], str):
            reqs_list = loads(str(types[0]))
            expr.__def_type__ = Callable[  # type: ignore[attr-defined]
                [
                    (_query_to_types(transform(r)) if r is not None else Any)
                    for r in reqs_list
                ]
            ]
        else:
            expr.__def_type__ = Callable[types]  # type: ignore[attr-defined]
        return expr

    return decorator


@define(Any, Any, Any, _name="each")
def each(var: DotHandler, col: DotHandler, expr: DotHandler) -> None:
    """Returns a list of the results of the expression evaluation for every
    item of the sequence provided.

    Example:

    .. code-block:: python

        each(S.x, S.collection, S.x.name)

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        [x.name for x in collection]

    """


@define(Any, Any, Any, _name="if")
def if_(test: t.Any, then: t.Any, else_: t.Any) -> None:
    """Checks condition and continues to evaluate one of the two expressions
    provided.

    Example:

    .. code-block:: python

        if_(S.value, 'truish', 'falsish')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if value:
            return 'truish'
        else:
            return 'falsish'

    """


@define(Any, Any, Any, _name="if_some")
def if_some(bind: t.List, then: _Func, else_: t.Any) -> None:
    """Used to unpack values with ``Optional`` types and using them safely in
    expressions.

    Example:

    .. code-block:: python

        if_some([S.img, S.this.image],
                image_url(S.img),
                'http://example.com/no-photo.jpg')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if this.image is not None:
            img = this.image
            return image_url(img)
        else:
            return 'http://example.com/no-photo.jpg'

    If ``S.this.image`` has a type ``Optional[TypeRef['Image']]``, ``S.img``
    variable will have a type ``TypeRef['Image']`` and it will be available only
    in "then"-expression, which will be evaluated only if ``S.this.image``
    wouldn't be ``None``. Otherwise "else"-expression will be
    evaluated.
    """



from typing import List

from hiku.graph import Graph, Node

from hiku.federation.directive import FieldSet, Key


def get_keys(graph: Graph, typename: str) -> List[FieldSet]:
    """Get all 'key' directives fields"""
    node = graph.nodes_map[typename]
    return [d.fields for d in node.directives if isinstance(d, Key)]


def get_entity_types(nodes: List[Node], federation_version: int) -> List[str]:
    entity_nodes = set()
    for node in nodes:
        if node.name is not None:
            for directive in node.directives:
                if isinstance(directive, Key):
                    if not directive.resolvable and federation_version == 2:
                        continue
                    entity_nodes.add(node.name)

    return list(sorted(entity_nodes))



import typing as t

from abc import abstractmethod, ABC
from collections import OrderedDict
from typing import TypeVar

from hiku.scalar import ScalarMeta

if t.TYPE_CHECKING:
    from hiku.graph import Union, Interface
    from hiku.enum import BaseEnum


class GenericMeta(type):
    def __repr__(cls) -> str:
        return cls.__name__

    def __eq__(cls, other: t.Any) -> bool:
        return (
            cls.__class__ is other.__class__ and cls.__dict__ == other.__dict__
        )

    def __ne__(cls, other: t.Any) -> bool:
        return not cls.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.__name__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        raise NotImplementedError(type(cls))


class AnyMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_any(cls)


class Any(metaclass=AnyMeta):
    pass


class BooleanMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_boolean(cls)


class Boolean(metaclass=BooleanMeta):
    pass


class StringMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_string(cls)


class String(metaclass=StringMeta):
    pass


class IDMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_id(cls)


class ID(metaclass=IDMeta):
    pass


class IntegerMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_integer(cls)


class Integer(metaclass=IntegerMeta):
    pass


class FloatMeta(GenericMeta):
    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_float(cls)


class Float(metaclass=FloatMeta):
    pass


TM = t.TypeVar("TM", bound="TypingMeta")


class TypingMeta(GenericMeta, type):
    __final__ = False

    def __cls_init__(cls: TM, parameters: t.Any) -> None:
        raise NotImplementedError(type(cls))

    def __cls_repr__(cls: TM) -> str:
        raise NotImplementedError(type(cls))

    def __getitem__(cls: TM, parameters: t.Any) -> TM:
        if cls.__final__:
            raise TypeError("Cannot substitute parameters in {!r}".format(cls))
        type_ = cls.__class__(cls.__name__, cls.__bases__, dict(cls.__dict__))
        type_.__cls_init__(parameters)
        type_.__final__ = True
        return type_

    def __repr__(self) -> str:
        if self.__final__:
            return self.__cls_repr__()
        else:
            return super(TypingMeta, self).__repr__()

    def __hash__(self) -> int:
        return hash(self.__name__)


class OptionalMeta(TypingMeta):
    __type__: GenericMeta

    def __cls_init__(cls, type_: GenericMeta) -> None:
        cls.__type__: GenericMeta = _maybe_typeref(type_)

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__type__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_optional(cls)


class Optional(metaclass=OptionalMeta):
    pass


class SequenceMeta(TypingMeta):
    __item_type__: GenericMeta

    def __cls_init__(cls, item_type: GenericMeta) -> None:
        cls.__item_type__: GenericMeta = _maybe_typeref(item_type)

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__item_type__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_sequence(cls)


class Sequence(metaclass=SequenceMeta):
    pass


class MappingMeta(TypingMeta):
    __key_type__: GenericMeta
    __value_type__: GenericMeta

    def __cls_init__(cls, params: t.Tuple[GenericMeta, GenericMeta]) -> None:
        key_type, value_type = params
        cls.__key_type__ = _maybe_typeref(key_type)
        cls.__value_type__ = _maybe_typeref(value_type)

    def __cls_repr__(self) -> str:
        return "{}[{!r}, {!r}]".format(
            self.__name__, self.__key_type__, self.__value_type__
        )

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_mapping(cls)


class Mapping(metaclass=MappingMeta):
    pass


class RecordMeta(TypingMeta):
    __field_types__: OrderedDict

    def __cls_init__(
        cls,
        field_types: t.Union[
            t.Dict[str, GenericMeta], t.List[t.Tuple[str, GenericMeta]]
        ],
    ) -> None:
        items: t.Iterable
        if hasattr(field_types, "items"):
            field_types = t.cast(t.Dict[str, GenericMeta], field_types)
            items = list(field_types.items())
        else:
            items = list(field_types)
        cls.__field_types__ = OrderedDict(
            (key, _maybe_typeref(val)) for key, val in items
        )

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, dict(self.__field_types__))

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_record(cls)


class Record(metaclass=RecordMeta):
    __field_types__: OrderedDict


class CallableMeta(TypingMeta):
    def __cls_init__(cls, arg_types: t.Iterable[GenericMeta]) -> None:
        cls.__arg_types__ = [_maybe_typeref(typ) for typ in arg_types]

    def __cls_repr__(self) -> str:
        return "{}[{}]".format(
            self.__name__, ", ".join(map(repr, self.__arg_types__))
        )

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_callable(cls)


class Callable(metaclass=CallableMeta):
    pass


class TypeRefMeta(TypingMeta):
    __type_name__: str

    def __cls_init__(cls, *args: str) -> None:
        assert len(args) == 1, f"{cls.__name__} takes exactly one argument"

        cls.__type_name__ = args[0]

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__type_name__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_typeref(cls)


class TypeRef(metaclass=TypeRefMeta):
    ...


class UnionRefMeta(TypingMeta):
    __type_name__: str

    def __cls_init__(cls, *args: str) -> None:
        assert len(args) == 1, f"{cls.__name__} takes exactly one argument"

        cls.__type_name__ = args[0]

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__type_name__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_unionref(cls)


class UnionRef(metaclass=UnionRefMeta):
    ...


class InterfaceRefMeta(TypingMeta):
    __type_name__: str

    def __cls_init__(cls, *args: str) -> None:
        assert len(args) == 1, f"{cls.__name__} takes exactly one argument"

        cls.__type_name__ = args[0]

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__type_name__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_interfaceref(cls)


class InterfaceRef(metaclass=InterfaceRefMeta):
    ...


class EnumRefMeta(TypingMeta):
    __type_name__: str

    def __cls_init__(cls, *args: str) -> None:
        assert len(args) == 1, f"{cls.__name__} takes exactly one argument"

        cls.__type_name__ = args[0]

    def __cls_repr__(self) -> str:
        return "{}[{!r}]".format(self.__name__, self.__type_name__)

    def accept(cls, visitor: "AbstractTypeVisitor") -> t.Any:
        return visitor.visit_enumref(cls)


class EnumRef(metaclass=EnumRefMeta):
    ...


RefMeta = (TypeRefMeta, UnionRefMeta, InterfaceRefMeta, EnumRefMeta)


@t.overload
def _maybe_typeref(typ: str) -> TypeRefMeta:
    ...


@t.overload
def _maybe_typeref(typ: GenericMeta) -> GenericMeta:
    ...


def _maybe_typeref(typ: t.Union[str, GenericMeta]) -> GenericMeta:
    return TypeRef[typ] if isinstance(typ, str) else typ


class AbstractTypeVisitor(ABC):
    def visit(self, obj: GenericMeta) -> t.Any:
        return obj.accept(self)

    @abstractmethod
    def visit_any(self, obj: AnyMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_boolean(self, obj: BooleanMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_string(self, obj: StringMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_id(self, obj: IDMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_integer(self, obj: IntegerMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_float(self, obj: FloatMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_typeref(self, obj: TypeRefMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_optional(self, obj: OptionalMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_sequence(self, obj: SequenceMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_mapping(self, obj: MappingMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_record(self, obj: RecordMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_callable(self, obj: CallableMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_unionref(self, obj: UnionRefMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_interfaceref(self, obj: InterfaceRefMeta) -> t.Any:
        pass

    @abstractmethod
    def visit_enumref(self, obj: EnumRefMeta) -> t.Any:
        pass


class TypeVisitor(AbstractTypeVisitor):
    def visit_any(self, obj: AnyMeta) -> t.Any:
        pass

    def visit_boolean(self, obj: BooleanMeta) -> t.Any:
        pass

    def visit_string(self, obj: StringMeta) -> t.Any:
        pass

    def visit_id(self, obj: IDMeta) -> t.Any:
        pass

    def visit_integer(self, obj: IntegerMeta) -> t.Any:
        pass

    def visit_float(self, obj: FloatMeta) -> t.Any:
        pass

    def visit_typeref(self, obj: TypeRefMeta) -> t.Any:
        pass

    def visit_unionref(self, obj: UnionRefMeta) -> t.Any:
        pass

    def visit_enumref(self, obj: EnumRefMeta) -> t.Any:
        pass

    def visit_optional(self, obj: OptionalMeta) -> t.Any:
        self.visit(obj.__type__)

    def visit_sequence(self, obj: SequenceMeta) -> t.Any:
        self.visit(obj.__item_type__)

    def visit_mapping(self, obj: MappingMeta) -> t.Any:
        self.visit(obj.__key_type__)
        self.visit(obj.__value_type__)

    def visit_record(self, obj: RecordMeta) -> t.Any:
        for value_type in obj.__field_types__.values():
            self.visit(value_type)

    def visit_callable(self, obj: CallableMeta) -> t.Any:
        for arg_type in obj.__arg_types__:
            self.visit(arg_type)


T = TypeVar("T", bound=GenericMeta)

Types = t.Mapping[str, t.Union[t.Type[Record], "Union", "Interface"]]


@t.overload
def get_type(  # type: ignore[misc]
    types: Types, typ: TypeRefMeta
) -> t.Type[Record]:
    ...


@t.overload
def get_type(types: Types, typ: UnionRefMeta) -> "Union":  # type: ignore[misc]
    ...


@t.overload
def get_type(  # type: ignore[misc]
    types: Types, typ: EnumRefMeta
) -> "BaseEnum":
    ...


@t.overload
def get_type(  # type: ignore[misc]
    types: Types, typ: InterfaceRefMeta
) -> "Interface":
    ...


@t.overload
def get_type(types: Types, typ: ScalarMeta) -> "ScalarMeta":
    ...


@t.overload
def get_type(types: Types, typ: T) -> T:
    ...


def get_type(types: Types, typ: t.Any) -> t.Any:
    if isinstance(typ, RefMeta):
        return types[typ.__type_name__]
    else:
        return typ



from concurrent.futures import (
    wait,
    FIRST_COMPLETED,
    Executor,
    Future,
)
from typing import (
    TYPE_CHECKING,
    Callable,
    Any,
)

from hiku.executors.base import BaseSyncExecutor
from hiku.result import Proxy

if TYPE_CHECKING:
    from hiku.executors.queue import (
        Queue,
        Workflow,
    )


class ThreadsExecutor(BaseSyncExecutor):
    def __init__(self, pool: Executor):
        self._pool = pool

    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> Future:
        return self._pool.submit(fn, *args, **kwargs)

    def process(self, queue: "Queue", workflow: "Workflow") -> Proxy:
        while queue.__futures__:
            done, _ = wait(  # type: ignore
                queue.__futures__, return_when=FIRST_COMPLETED  # type: ignore
            )
            queue.progress(done)
        return workflow.result()



"""
    hiku.query
    ~~~~~~~~~~

    `Hiku` doesn't rely on any specific query language, internally it uses
    generic query representation to describe result and it could be constructed
    by parsing different suitable query languages.

    However, `Hiku` provides one built-in way to describe result -- `edn`_ data
    structure -- `simple` queries, which are very similar to the `om.next`_
    queries, which are inspired by `Datomic Pull API`_.

      - ``[:foo]`` - node fields definition (`edn` keywords in vector)
      - ``{:bar [:baz]}`` - link definition (`edn` map with keyword and vector)
      - ``(:foo {:key val})`` - field or link options definition (field name or
        link name, wrapped with `edn` list with map of options as a second
        value)

    Example:

    .. code-block:: clojure

        [:foo {:bar [:baz]}]

    This query will be read internally as:

    .. code-block:: python

        Node([Field('foo'),
              Link('bar', Node([Field('baz')]))])

    And query result will look like this:

    .. code-block:: python

        {
            'foo': 1,
            'bar': {
                'baz': 2,
            },
        }

    .. _edn: https://github.com/edn-format/edn
    .. _Datomic Pull API: http://docs.datomic.com/pull.html
    .. _om.next: https://github.com/omcljs/om/wiki/Documentation-(om.next)
"""
import typing as t
import hashlib

from itertools import chain
from collections import (
    OrderedDict,
    defaultdict,
)
from collections.abc import Sequence

from typing_extensions import TypeAlias

from .directives import Directive
from .utils import cached_property

T = t.TypeVar("T", bound="Base")


def _compute_hash(obj: t.Any) -> int:
    if isinstance(obj, dict):
        return hash(
            tuple(
                (_compute_hash(k), _compute_hash(v))
                for k, v in sorted(obj.items())
            )
        )
    if isinstance(obj, list):
        return hash(tuple(_compute_hash(i) for i in obj))
    if isinstance(obj, bytes):
        return int(hashlib.sha1(obj).hexdigest(), 16)
    if isinstance(obj, str):
        return int(hashlib.sha1(obj.encode("utf-8")).hexdigest(), 16)
    return hash(obj)


class Base:
    __attrs__: t.Tuple[str, ...] = ()

    def __repr__(self) -> str:
        kwargs = ", ".join(
            "{}={!r}".format(attr, self.__dict__[attr])
            for attr in self.__attrs__
        )
        return "{}({})".format(self.__class__.__name__, kwargs)

    def __eq__(self, other: t.Any) -> bool:
        return self.__class__ is other.__class__ and all(
            self.__dict__[attr] == other.__dict__[attr]
            for attr in self.__attrs__
        )

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def copy(self: T, **kwargs: t.Any) -> T:
        obj = self.__class__.__new__(self.__class__)
        obj.__dict__.update(
            (attr, kwargs.get(attr, self.__dict__[attr]))
            for attr in self.__attrs__
        )
        return obj


class FieldBase(Base):
    name: str
    options: t.Optional[t.Dict[str, t.Any]]
    alias: t.Optional[str]

    @cached_property
    def result_key(self) -> str:
        if self.alias is not None:
            return self.alias
        else:
            return self.name

    @cached_property
    def options_hash(self) -> t.Optional[int]:
        if self.options:
            return _compute_hash(self.options)
        else:
            return None

    @cached_property
    def index_key(self) -> str:
        if self.options_hash is not None:
            return "{}[{}]".format(self.name, self.options_hash)
        else:
            return self.name

    # TODO: test this hash
    def __hash__(self) -> int:
        return hash(self.index_key)


class Field(FieldBase):
    """Represents a field of the node

    :param name: name of the field
    :param optional options: field options -- mapping of names to values
    :param optional alias: field's name in result
    """

    __attrs__ = ("name", "options", "alias", "directives")

    def __init__(
        self,
        name: str,
        options: t.Optional[t.Dict[str, t.Any]] = None,
        alias: t.Optional[str] = None,
        directives: t.Optional[t.Tuple[Directive, ...]] = None,
    ):
        self.name = name
        self.options = options
        self.alias = alias
        self.directives = directives or ()

    @cached_property
    def directives_map(self) -> OrderedDict:
        return OrderedDict(
            (d.__directive_info__.name, d) for d in self.directives
        )

    def accept(self, visitor: "QueryVisitor") -> t.Any:
        return visitor.visit_field(self)


class Link(FieldBase):
    """Represents a link to the node

    :param name: name of the link
    :param node: collection of fields and links --
                 :py:class:`~hiku.query.Node`
    :param optional options: link options -- mapping of names to values
    :param optional alias: link's name in result
    """

    __attrs__ = (
        "name",
        "node",
        "options",
        "alias",
        "directives",
    )

    def __init__(
        self,
        name: str,
        node: "Node",
        options: t.Optional[t.Dict[str, t.Any]] = None,
        alias: t.Optional[str] = None,
        directives: t.Optional[t.Tuple[Directive, ...]] = None,
    ):
        self.name = name
        self.node = node
        self.options = options
        self.alias = alias
        self.directives = directives or ()

    @cached_property
    def directives_map(self) -> OrderedDict:
        return OrderedDict(
            (d.__directive_info__.name, d) for d in self.directives
        )

    def accept(self, visitor: "QueryVisitor") -> t.Any:
        return visitor.visit_link(self)


FieldOrLink = t.Union[Field, Link]
FieldsMap: TypeAlias = "OrderedDict[str, FieldBase]"
FragmentMap: TypeAlias = "OrderedDict[str, Fragment]"


class Node(Base):
    """Represents collection of fields, links and fragments

    :param fields: list of :py:class:`~hiku.query.Field`,
        :py:class:`~hiku.query.Link`, :py:class:`~hiku.query.Fragment`
    :param fragments: list of :py:class:`~hiku.query.Fragment`
    :param ordered: whether to compute fields of this node sequentially
        in order or not
    """

    __attrs__ = ("fields", "fragments", "ordered")

    def __init__(
        self,
        fields: t.Sequence[FieldOrLink],
        fragments: t.Optional[t.Sequence["Fragment"]] = None,
        ordered: bool = False,
    ) -> None:
        self.fields = list(fields)
        self.fragments = list(fragments or [])
        self.ordered = ordered

    @cached_property
    def fields_map(
        self,
    ) -> FieldsMap:
        return OrderedDict((f.name, f) for f in self.fields)

    @cached_property
    def fragments_map(self) -> FragmentMap:
        return OrderedDict((f.type_name, f) for f in self.fragments)

    @cached_property
    def result_map(self) -> OrderedDict:
        return OrderedDict((f.result_key, f) for f in self.fields)

    def accept(self, visitor: "QueryVisitor") -> t.Any:
        return visitor.visit_node(self)

    def __hash__(self) -> int:
        return hash(tuple(self.fields)) + hash(tuple(self.fragments))


class Fragment(Base):
    __attrs__ = ("type_name", "node")

    def __init__(self, type_name: str, fields: t.List[FieldOrLink]) -> None:
        self.type_name = type_name
        self.node = Node(fields)

    def accept(self, visitor: "QueryVisitor") -> t.Any:
        return visitor.visit_fragment(self)

    def __hash__(self) -> int:
        return hash(self.type_name) + hash(self.node)


KeyT = t.Tuple[str, t.Optional[str], t.Optional[str]]


def field_key(field: FieldOrLink) -> KeyT:
    return (field.name, field.options_hash, field.alias)


def _merge(
    nodes: t.Iterable[Node],
) -> t.Iterator[t.Union[FieldOrLink, Fragment]]:
    visited_fields = set()
    visited_fragments = set()
    links = {}
    link_directives: t.DefaultDict[t.Tuple, t.List] = defaultdict(list)
    to_merge = OrderedDict()
    fields_iter = chain.from_iterable(e.fields for e in nodes)
    fragments_iter = chain.from_iterable(e.fragments for e in nodes)

    for field in fields_iter:
        key = field_key(field)

        if field.__class__ is Link:
            field = t.cast(Link, field)
            if key not in to_merge:
                to_merge[key] = [field.node]
                links[key] = field
            else:
                to_merge[key].append(field.node)
            link_directives[key].extend(field.directives)
        else:
            if key not in visited_fields:
                visited_fields.add(key)
                yield field

    if not visited_fields and not to_merge:
        for fr in fragments_iter:
            yield fr
    else:
        for fr in fragments_iter:
            fr_fields = []
            for field in fr.node.fields:
                key = (field.name, field.options_hash, field.alias)

                if field.__class__ is Link:
                    field = t.cast(Link, field)
                    if key not in to_merge:
                        to_merge[key] = [field.node]
                        links[key] = field
                    else:
                        to_merge[key].append(field.node)
                    link_directives[key].extend(field.directives)
                else:
                    if key not in visited_fields:
                        fr_fields.append(field)

            fr_key = (fr.type_name, tuple(field_key(f) for f in fr_fields))
            if fr_key not in visited_fragments:
                visited_fragments.add(fr_key)
                if fr_fields:
                    yield Fragment(fr.type_name, fr_fields)

    for key, values in to_merge.items():
        link = links[key]
        directives = link_directives[key]
        yield link.copy(node=merge(values), directives=tuple(directives))


def merge(nodes: t.Iterable[Node]) -> Node:
    """Merges multiple queries into one query

    :param nodes: queries, represented as list of :py:class:`~hiku.query.Node`
    :return: merged query as one :py:class:`~hiku.query.Node`
    """
    assert isinstance(nodes, Sequence), type(nodes)
    ordered = any(n.ordered for n in nodes)
    fields = []
    fragments = []
    for item in _merge(nodes):
        if isinstance(item, Fragment):
            fragments.append(item)
        else:
            fields.append(item)

    return Node(fields=fields, fragments=fragments, ordered=ordered)


class QueryVisitor:
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_field(self, obj: Field) -> t.Any:
        pass

    def visit_link(self, obj: Link) -> t.Any:
        self.visit(obj.node)

    def visit_node(self, obj: Node) -> t.Any:
        for item in obj.fields:
            self.visit(item)

    def visit_fragment(self, obj: Fragment) -> t.Any:
        self.visit(obj.node)


class QueryTransformer:
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_field(self, obj: Field) -> Field:
        return obj.copy()

    def visit_link(self, obj: Link) -> Link:
        return obj.copy(node=self.visit(obj.node))

    def visit_node(self, obj: Node) -> Node:
        return obj.copy(fields=[self.visit(f) for f in obj.fields])

    def visit_fragment(self, obj: Fragment) -> Fragment:
        return obj.copy(node=self.visit(obj.node))



import typing as t

from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.graph import Graph
from hiku.result import Proxy
from hiku.types import Record, Sequence, TypeRef


class DenormalizeEntityGraphQL(DenormalizeGraphQL):
    def __init__(
        self, graph: Graph, result: Proxy, entity_type_name: str
    ) -> None:
        super().__init__(graph, result, "Query")
        t.cast(Record, self._type[-1]).__field_types__["_entities"] = Sequence[
            TypeRef[entity_type_name]
        ]



import typing as t


VisitorT = t.Union["NodeTransformer", "NodeVisitor"]


class Node:
    def accept(self, visitor: VisitorT) -> t.Any:
        raise NotImplementedError(type(self))


class Symbol(Node):
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return self.name

    def accept(self, visitor: VisitorT) -> t.Any:
        return visitor.visit_symbol(self)


class Keyword(Node):
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return ":{}".format(self.name)

    def accept(self, visitor: VisitorT) -> t.Any:
        return visitor.visit_keyword(self)


class Tuple(Node):
    def __init__(self, values: t.List[Node]) -> None:
        self.values = tuple(values)

    def __repr__(self) -> str:
        return "({})".format(" ".join(map(repr, self.values)))

    def accept(self, visitor: VisitorT) -> t.Any:
        return visitor.visit_tuple(self)


class List(Node):
    def __init__(self, values: t.List[Node]) -> None:
        self.values = tuple(values)

    def __repr__(self) -> str:
        return "[{}]".format(" ".join(map(repr, self.values)))

    def accept(self, visitor: VisitorT) -> t.Any:
        return visitor.visit_list(self)


class Dict(Node):
    def __init__(self, values: t.List[Node]):
        self.values = tuple(values)

    def __repr__(self) -> str:
        return "{{{}}}".format(" ".join(map(repr, self.values)))

    def accept(self, visitor: VisitorT) -> t.Any:
        return visitor.visit_dict(self)


class NodeVisitor:
    def visit(self, node: Node) -> t.Any:
        if hasattr(node, "accept"):
            node.accept(self)
        else:
            self.generic_visit(node)

    def generic_visit(self, node: Node) -> t.Any:
        pass

    def visit_symbol(self, node: Symbol) -> t.Any:
        pass

    def visit_keyword(self, node: Keyword) -> t.Any:
        pass

    def visit_tuple(self, node: Tuple) -> t.Any:
        for value in node.values:
            self.visit(value)

    def visit_list(self, node: List) -> t.Any:
        for value in node.values:
            self.visit(value)

    def visit_dict(self, node: Dict) -> t.Any:
        for value in node.values:
            self.visit(value)


class NodeTransformer:
    def visit(self, node: Node) -> t.Any:
        if hasattr(node, "accept"):
            return node.accept(self)

        return self.generic_visit(node)

    def generic_visit(self, node: Node) -> Node:
        return node

    def visit_symbol(self, node: Symbol) -> Symbol:
        return Symbol(node.name)

    def visit_keyword(self, node: Keyword) -> Keyword:
        return Keyword(node.name)

    def visit_tuple(self, node: Tuple) -> Tuple:
        return Tuple([self.visit(value) for value in node.values])

    def visit_list(self, node: List) -> List:
        return List([self.visit(value) for value in node.values])

    def visit_dict(self, node: Dict) -> Dict:
        return Dict([self.visit(value) for value in node.values])



from contextlib import contextmanager
from collections import deque

from ..types import Sequence, SequenceMeta, Record, RecordMeta, get_type
from ..types import MappingMeta, OptionalMeta, Any, AnyMeta

from .refs import NamedRef, Ref
from .nodes import NodeTransformer, Symbol, Keyword, Tuple, List


def fn_types(functions):
    return {fn.__def_name__: fn.__def_type__ for fn in functions}


class Environ:
    def __init__(self, values):
        env_vars = {}
        for name, value in values.items():
            env_vars[name] = NamedRef(None, name, value)
        self.vars = deque([env_vars])

    @contextmanager
    def push(self, mapping):
        self.vars.append(mapping)
        try:
            yield
        finally:
            self.vars.pop()

    def __getitem__(self, key):
        for d in reversed(self.vars):
            try:
                return d[key]
            except KeyError:
                continue
        else:
            raise KeyError(repr(key))

    def __contains__(self, key):
        return any(key in d for d in self.vars)


def node_type(types, node):
    return get_type(types, node.__ref__.to)


def check_type(types, t1, t2):
    t1 = get_type(types, t1)
    t2 = get_type(types, t2)
    if isinstance(t2, AnyMeta):
        pass
    else:
        if isinstance(t1, type(t2)):
            if isinstance(t2, OptionalMeta):
                check_type(types, t1.__type__, t2.__type__)
            elif isinstance(t2, SequenceMeta):
                check_type(types, t1.__item_type__, t2.__item_type__)
            elif isinstance(t2, MappingMeta):
                check_type(types, t1.__key_type__, t2.__key_type__)
                check_type(types, t1.__value_type__, t2.__value_type__)
            elif isinstance(t2, RecordMeta):
                for key, v2 in t2.__field_types__.items():
                    v1 = t1.__field_types__.get(key)
                    if v1 is None:
                        raise TypeError('Missing field "{}"'.format(key))
                    v1 = get_type(types, v1)
                    check_type(types, v1, v2)
        elif isinstance(t2, OptionalMeta):
            check_type(types, t1, t2.__type__)
        else:
            raise TypeError("Types mismatch, {!r} != {!r}".format(t1, t2))


class Checker(NodeTransformer):
    def __init__(self, types, env):
        self.types = types
        self.env = Environ(env)

    def visit_get_expr(self, node):
        sym, obj, name = node.values
        assert isinstance(name, Symbol), type(name)
        obj = self.visit(obj)
        assert hasattr(obj, "__ref__"), "Object does not have a reference"

        ref_to = node_type(self.types, obj)
        check_type(self.types, ref_to, Record[{name.name: Any}])

        res = ref_to.__field_types__.get(name.name)
        assert res is not None, "Undefined field name: {}".format(name.name)
        tup = Tuple([sym, obj, name])
        tup.__ref__ = NamedRef(obj.__ref__, name.name, res)
        return tup

    def visit_each_expr(self, node):
        var, col, expr = node.values[1:]
        assert isinstance(var, Symbol), repr(var)
        col = self.visit(col)
        assert hasattr(col, "__ref__"), "Object does not have a reference"
        col_type = node_type(self.types, col)
        check_type(self.types, col_type, Sequence[Any])
        var = Symbol(var.name)
        var.__ref__ = Ref(col.__ref__, col_type.__item_type__)
        with self.env.push({var.name: var.__ref__}):
            expr = self.visit(expr)
        return Tuple([Symbol("each"), var, col, expr])

    def visit_if_some_expr(self, node):
        bind, then, else_ = node.values[1:]
        assert isinstance(bind, List) and len(bind.values) == 2, repr(bind)
        bind_sym, bind_expr = bind.values
        assert isinstance(bind_sym, Symbol), repr(bind_sym)
        bind_expr = self.visit(bind_expr)
        assert hasattr(bind_expr, "__ref__"), "Object does not have a reference"
        bind_type = node_type(self.types, bind_expr)
        if isinstance(bind_type, OptionalMeta):
            bind_expr_ref = Ref(bind_expr.__ref__, bind_type.__type__)
        else:
            # TODO: warn about unnecessary check
            bind_expr_ref = bind_expr.__ref__
        bind_sym = Symbol(bind_sym.name)
        bind_sym.__ref__ = bind_expr_ref
        with self.env.push({bind_sym.name: bind_sym.__ref__}):
            then = self.visit(then)
        else_ = self.visit(else_)
        return Tuple(
            [Symbol("if_some"), List([bind_sym, bind_expr]), then, else_]
        )

    def visit_tuple_generic(self, node):
        sym = self.visit(node.values[0])
        assert isinstance(sym, Symbol), type(sym)
        args = [self.visit(val) for val in node.values[1:]]
        fn_type = node_type(self.types, sym)
        assert len(fn_type.__arg_types__) == len(
            args
        ), 'Wrong arguments count at "{}" function'.format(sym.name)
        for arg, arg_type in zip(args, fn_type.__arg_types__):
            ref = getattr(arg, "__ref__", None)
            if ref is not None:
                check_type(self.types, node_type(self.types, arg), arg_type)
            else:
                check_type(self.types, Any, arg_type)
        return Tuple([sym] + args)

    def visit_tuple(self, node):
        sym = node.values[0]
        if sym.name == "get":
            return self.visit_get_expr(node)
        elif sym.name == "each":
            return self.visit_each_expr(node)
        elif sym.name == "if_some":
            return self.visit_if_some_expr(node)
        else:
            return self.visit_tuple_generic(node)

    def visit_symbol(self, node):
        if node.name not in self.env:
            raise TypeError('Unknown symbol "{}"'.format(node.name))
        sym = Symbol(node.name)
        sym.__ref__ = self.env[node.name]
        return sym

    def visit_dict(self, node):
        assert not len(node.values) % 2, "Probably missing keyword value"
        keys = node.values[::2]
        assert all(isinstance(k, Keyword) for k in keys), "Wrong arguments"
        return super(Checker, self).visit_dict(node)


def check(expr, types, env):
    return Checker(types, env).visit(expr)



from collections import defaultdict

from typing import (
    Optional,
    Dict,
    Iterator,
    Tuple,
    Union,
    List,
    cast,
    Any,
    Set,
)

from graphql.language import ast
from graphql.language.parser import parse
from hiku.utils import ImmutableDict

from ..directives import (
    Cached,
    Directive,
)
from ..operation import Operation, OperationType
from ..query import FieldOrLink, Fragment, Node, Field, Link, merge


def parse_query(src: str) -> ast.DocumentNode:
    """Parses a query into GraphQL ast

    :param str src: GraphQL query string
    :return: :py:class:`ast.DocumentNode`
    """
    return parse(src)


class NodeVisitor:
    def visit(self, obj: ast.Node) -> Any:
        visit_method = getattr(self, "visit_{}".format(obj.kind))
        if visit_method is None:
            raise NotImplementedError(
                "Not implemented node type: {!r}".format(obj)
            )
        return visit_method(obj)

    def visit_document(self, obj: ast.DocumentNode) -> None:
        for definition in obj.definitions:
            self.visit(definition)

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> Any:
        self.visit(obj.selection_set)

    def visit_fragment_definition(self, obj: ast.FragmentDefinitionNode) -> Any:
        self.visit(obj.selection_set)

    def visit_selection_set(self, obj: ast.SelectionSetNode) -> Any:
        for i in obj.selections:
            self.visit(i)

    def visit_field(self, obj: ast.FieldNode) -> Any:
        pass

    def visit_fragment_spread(self, obj: ast.FragmentSpreadNode) -> Any:
        pass

    def visit_inline_fragment(self, obj: ast.InlineFragmentNode) -> Any:
        self.visit(obj.selection_set)


class OperationGetter(NodeVisitor):
    def __init__(self, operation_name: Optional[str] = None):
        self._operations: Dict[Optional[str], ast.OperationDefinitionNode] = {}
        self._operation_name = operation_name

    @classmethod
    def get(
        cls, doc: ast.DocumentNode, operation_name: Optional[str] = None
    ) -> ast.OperationDefinitionNode:
        self = cls(operation_name=operation_name)
        self.visit(doc)
        if not self._operations:
            raise TypeError("No operations in the document")

        if self._operation_name is None:
            if len(self._operations) > 1:
                raise TypeError(
                    "Document should contain exactly one operation "
                    "when no operation name was provided"
                )
            return next(iter(self._operations.values()))
        else:
            try:
                return self._operations[self._operation_name]
            except KeyError:
                raise ValueError(
                    "Undefined operation name: {!r}".format(
                        self._operation_name
                    )
                )

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> None:
        pass  # skip visit here

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        name = obj.name.value if obj.name is not None else None
        if name in self._operations:
            raise TypeError("Duplicate operation definition: {!r}".format(name))
        self._operations[name] = obj


class FragmentsCollector(NodeVisitor):
    def __init__(self) -> None:
        self.fragments_map: Dict[str, ast.FragmentDefinitionNode] = {}

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        pass  # not interested in operations here

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> None:
        if obj.name.value in self.fragments_map:
            raise TypeError(
                'Duplicated fragment name: "{}"'.format(obj.name.value)
            )
        self.fragments_map[obj.name.value] = obj


class SelectionSetVisitMixin:
    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        raise NotImplementedError(type(self))

    @property
    def query_variables(self) -> Optional[Dict]:
        raise NotImplementedError(type(self))

    @property
    def query_name(self) -> Optional[str]:
        raise NotImplementedError(type(self))

    def lookup_variable(self, name: str) -> Any:
        try:
            assert self.query_variables is not None
            return self.query_variables[name]
        except KeyError:
            raise TypeError(
                "Variable ${} is not defined in query {}".format(
                    name, self.query_name or "<unnamed>"
                )
            )

    def visit_selection_set(
        self, obj: ast.SelectionSetNode
    ) -> Iterator[Union[Field, Link]]:
        for i in obj.selections:
            for j in self.visit(i):  # type: ignore[attr-defined]
                yield j

    def _should_skip(self, obj: ast.SelectionNode) -> Optional[bool]:
        if not obj.directives:
            return None

        skip = next((d for d in obj.directives if d.name.value == "skip"), None)
        if skip is not None:
            if len(skip.arguments) != 1:
                raise TypeError(
                    "@skip directive accepts exactly one "
                    "argument, {} provided".format(len(skip.arguments))
                )
            skip_arg = skip.arguments[0]
            if skip_arg.name.value != "if":
                raise TypeError(
                    '@skip directive does not accept "{}" '
                    "argument".format(skip_arg.name.value)
                )
            return self.visit(skip_arg.value)  # type: ignore[attr-defined]

        include = next(
            (d for d in obj.directives if d.name.value == "include"), None
        )
        if include is not None:
            if len(include.arguments) != 1:
                raise TypeError(
                    "@include directive accepts exactly one "
                    "argument, {} provided".format(len(include.arguments))
                )
            include_arg = include.arguments[0]
            if include_arg.name.value != "if":
                raise TypeError(
                    '@include directive does not accept "{}" '
                    "argument".format(include_arg.name.value)
                )
            return not self.visit(include_arg.value)  # type: ignore[attr-defined] # noqa: E501

        return None

    def _get_directive(
        self, name: str, obj: ast.SelectionNode
    ) -> Optional[ast.DirectiveNode]:
        return next((d for d in obj.directives if d.name.value == name), None)

    def _is_cached(self, obj: ast.SelectionNode) -> Optional[Cached]:
        if not obj.directives:
            return None

        cached = self._get_directive("cached", obj)
        if cached is None:
            return None

        if len(cached.arguments) != 1:
            raise TypeError(
                "@cached directive accepts exactly one "
                "argument, {} provided".format(len(cached.arguments))
            )
        cached_arg = cached.arguments[0]
        if cached_arg.name.value != "ttl":
            raise TypeError(
                '@cached directive does not accept "{}" '
                "argument".format(cached_arg.name.value)
            )
        ttl = self.visit(cached_arg.value)  # type: ignore[attr-defined]
        if not isinstance(ttl, int):
            raise TypeError("@cached ttl argument must be an integer")

        return Cached(ttl=ttl)

    def _collect_fields(
        self, obj: ast.FieldNode
    ) -> Tuple[List[FieldOrLink], List[Fragment]]:
        """Collect fields from AST node.

        We collect shared fields and type-specific fragments separately.
        Fragments for same type are merged.
        Resulting list of fragments contains only type-specific fragments.
        """
        assert obj.selection_set is not None

        if isinstance(self, FragmentsTransformer):
            fragments_map = self.fragments_map
        else:
            fragments_map = self.fragments_transformer.fragments_map  # type: ignore[attr-defined] # noqa: E501

        shared_fields = []
        type_fields = defaultdict(list)

        for item in obj.selection_set.selections:
            type_name = None
            selection_set = None

            if isinstance(item, ast.InlineFragmentNode):
                type_name = item.type_condition.name.value
                selection_set = item.selection_set
            elif isinstance(item, ast.FragmentSpreadNode):
                fragment = fragments_map[item.name.value]
                type_name = fragment.type_condition.name.value
                selection_set = fragment.selection_set
            else:
                shared_fields.extend(list(self.visit(item)))  # type: ignore[attr-defined] # noqa: E501

            if type_name and selection_set:
                type_fields[type_name].extend(
                    list(self.visit(selection_set))  # type: ignore[attr-defined] # noqa: E501
                )

        node_fragments = []
        for type_name, fields in type_fields.items():
            node_fragments.append(Fragment(type_name, fields))

        return shared_fields, node_fragments

    def visit_field(self, obj: ast.FieldNode) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        directives: List[Directive] = []
        cached = self._is_cached(obj)
        if cached is not None:
            directives.append(cached)

        if obj.arguments:
            options = {
                arg.name.value: self.visit(arg.value)  # type: ignore[attr-defined] # noqa: E501
                for arg in obj.arguments
            }
        else:
            options = None

        if obj.alias is not None:
            alias = obj.alias.value
        else:
            alias = None

        if obj.selection_set is None:
            yield Field(
                obj.name.value,
                options=options,
                alias=alias,
                directives=tuple(directives),
            )
        else:
            fields, fragments = self._collect_fields(obj)

            yield Link(
                obj.name.value,
                Node(fields, fragments),
                options=options,
                alias=alias,
                directives=tuple(directives),
            )

    def visit_variable(self, obj: ast.VariableNode) -> Any:
        return self.lookup_variable(obj.name.value)

    def visit_null_value(self, obj: ast.NullValueNode) -> None:
        return None

    def visit_int_value(self, obj: ast.IntValueNode) -> int:
        return int(obj.value)

    def visit_float_value(self, obj: ast.FloatValueNode) -> float:
        return float(obj.value)

    def visit_string_value(self, obj: ast.StringValueNode) -> str:
        return obj.value

    def visit_boolean_value(self, obj: ast.BooleanValueNode) -> bool:
        return obj.value

    def visit_enum_value(self, obj: ast.EnumValueNode) -> str:
        return obj.value

    def visit_list_value(self, obj: ast.ListValueNode) -> List:
        return [self.visit(i) for i in obj.values]  # type: ignore[attr-defined]

    def visit_object_value(self, obj: ast.ObjectValueNode) -> Dict:
        return {f.name.value: self.visit(f.value) for f in obj.fields}  # type: ignore[attr-defined] # noqa: E501

    def visit_fragment_spread(
        self, obj: ast.FragmentSpreadNode
    ) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        for i in self.transform_fragment(obj.name.value):
            yield i

    def visit_inline_fragment(
        self, obj: ast.InlineFragmentNode
    ) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        for i in self.visit(obj.selection_set):  # type: ignore[attr-defined]
            yield i


class FragmentsTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name: str = ""
    query_variables: Dict = {}

    def __init__(
        self, document: ast.DocumentNode, query_name: str, query_variables: Dict
    ):
        collector = FragmentsCollector()
        collector.visit(document)
        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_map = collector.fragments_map
        self.cache: Dict[str, List[Union[Field, Link]]] = {}
        self.pending_fragments: Set[str] = set()

    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        return self.visit(self.fragments_map[name])

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        pass  # not interested in operations here

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> List[Union[Field, Link]]:
        if obj.name.value in self.cache:
            return self.cache[obj.name.value]
        else:
            if obj.name.value in self.pending_fragments:
                raise TypeError(
                    'Cyclic fragment usage: "{}"'.format(obj.name.value)
                )
            self.pending_fragments.add(obj.name.value)
            try:
                selection_set = list(self.visit(obj.selection_set))
            finally:
                self.pending_fragments.discard(obj.name.value)
            self.cache[obj.name.value] = selection_set
            return selection_set


class GraphQLTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name: Optional[str] = None
    query_variables: Optional[Dict[str, Any]] = None
    fragments_transformer = None

    def __init__(
        self, document: ast.DocumentNode, variables: Optional[Dict] = None
    ):
        self.document = document
        self.variables = variables

    @classmethod
    def transform(
        cls,
        document: ast.DocumentNode,
        op: ast.OperationDefinitionNode,
        variables: Optional[Dict] = None,
    ) -> Node:
        visitor = cls(document, variables)
        return visitor.visit(op)

    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        assert self.fragments_transformer
        return self.fragments_transformer.transform_fragment(name)

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> Node:
        variables = self.variables or {}
        query_name = obj.name.value if obj.name else "<unnamed>"
        query_variables = {}
        for var_defn in obj.variable_definitions or ():
            name = var_defn.variable.name.value
            try:
                value = variables[name]  # TODO: check variable type
            except KeyError:
                if var_defn.default_value is not None:
                    value = self.visit(var_defn.default_value)
                elif isinstance(var_defn.type, ast.NonNullTypeNode):
                    raise TypeError(
                        'Variable "{}" is not provided for query {}'.format(
                            name, query_name
                        )
                    )
                else:
                    value = None
            query_variables[name] = value

        self.query_name = query_name
        assert self.query_name is not None
        self.query_variables = query_variables
        self.fragments_transformer = FragmentsTransformer(
            self.document, self.query_name, self.query_variables
        )
        ordered = obj.operation is ast.OperationType.MUTATION
        try:
            node = Node(list(self.visit(obj.selection_set)), ordered=ordered)
        finally:
            self.query_name = None
            self.query_variables = None
            self.fragments_transformer = None
        return merge([node])


def get_operation(
    doc: ast.DocumentNode, operation_name: Optional[str]
) -> ast.OperationDefinitionNode:
    return OperationGetter.get(doc, operation_name=operation_name)


def read(
    src: str,
    variables: Optional[Dict] = None,
    operation_name: Optional[str] = None,
) -> Node:
    """Reads a query from the GraphQL document

    Example:

    .. code-block:: python

        query = read('{ foo bar }')
        result = engine.execute(graph, query)

    :param str src: GraphQL query
    :param dict variables: query variables
    :param str operation_name: Name of the operation to execute
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    doc = parse_query(src)
    op = get_operation(doc, operation_name=operation_name)
    if op.operation is not ast.OperationType.QUERY:
        raise TypeError(
            'Only "query" operations are supported, '
            '"{}" operation was provided'.format(op.operation.value)
        )

    return GraphQLTransformer.transform(doc, op, variables)


def read_operation(
    src: Union[str, ast.DocumentNode],
    variables: Optional[Union[Dict, ImmutableDict]] = None,
    operation_name: Optional[str] = None,
) -> Operation:
    """Reads an operation from the GraphQL document

    Example:

    .. code-block:: python

        op = read_operation('{ foo bar }')
        if op.type is OperationType.QUERY:
            result = engine.execute(op.query, query_graph)

    :return: :py:class:`Operation`
    """
    if isinstance(src, str):
        doc = parse_query(src)
    elif isinstance(src, ast.DocumentNode):
        doc = src
    else:
        raise TypeError("Unsupported type: {}".format(type(src)))

    op = get_operation(doc, operation_name=operation_name)
    query = GraphQLTransformer.transform(doc, op, variables)
    type_ = cast(
        Optional[OperationType],
        (OperationType._value2member_map_.get(op.operation)),
    )
    name = op.name.value if op.name else None
    if type_ is None:
        raise TypeError("Unsupported operation type: {}".format(op.operation))

    return Operation(type_, query, name)



from collections import deque

from ..graph import Graph, Interface, Union
from ..query import Field, Link
from ..result import Proxy
from ..types import (
    RecordMeta,
    RefMeta,
    SequenceMeta,
    OptionalMeta,
    GenericMeta,
)

from .base import Denormalize


class DenormalizeGraphQL(Denormalize):
    def __init__(
        self, graph: Graph, result: Proxy, root_type_name: str
    ) -> None:
        super().__init__(graph, result)
        self._type_name = deque([root_type_name])

    def visit_field(self, obj: Field) -> None:
        if obj.name == "__typename":
            type_name = self._type_name[-1]
            if isinstance(self._type[-1], (Union, Interface)):
                type_name = self._data[-1].__ref__.node
            self._res[-1][obj.result_key] = type_name
        else:
            super().visit_field(obj)

    def visit_link(self, obj: Link) -> None:
        if isinstance(self._type[-1], (Union, Interface)):
            type_ = self._types[self._data[-1].__ref__.node].__field_types__[
                obj.name
            ]
        elif isinstance(self._type[-1], RecordMeta):
            type_ = self._type[-1].__field_types__[obj.name]
        else:
            raise AssertionError(repr(self._type[-1]))

        type_ref: GenericMeta
        if isinstance(type_, RefMeta):
            type_ref = type_
        elif isinstance(type_, SequenceMeta):
            type_ref = type_.__item_type__
            if isinstance(type_ref, OptionalMeta):
                type_ref = type_ref.__type__
            assert isinstance(type_ref, RefMeta), type_ref
        elif isinstance(type_, OptionalMeta):
            type_ref = type_.__type__
            assert isinstance(type_ref, RefMeta), type_ref
        else:
            raise AssertionError(repr(type_))
        self._type_name.append(type_ref.__type_name__)
        super().visit_link(obj)
        self._type_name.pop()



import uuid

from datetime import date, datetime
from typing import Any, Callable, Optional, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from hiku.graph import AbstractGraphVisitor


def scalar(
    name: Optional[str] = None, description: Optional[str] = None
) -> Callable[[Type["Scalar"]], Type["Scalar"]]:
    """
    Use @scalar decorator to set custom name and description for
    the scalar type.

    It is not mandatory to use this decorator, as long as
    you do not need to set custom name and description for the scalar type.
    """

    def _scalar(cls: Type["Scalar"]) -> Type["Scalar"]:
        cls.__type_name__ = name or cls.__name__
        cls.__description__ = description
        return cls

    return _scalar


class ScalarMeta(type):
    """Metaclass for scalar types.
    Handles `__type_name__` attribute.
    """

    __type_name__: str
    __description__: Optional[str]

    def __new__(cls, *args: Any, **kwargs: Any):  # type: ignore[no-untyped-def]
        instance = super().__new__(cls, *args, **kwargs)
        instance.__type_name__ = instance.__name__
        return instance


class Scalar(metaclass=ScalarMeta):
    """
    Base class for scalar types.

    To implement custom scalar type, subclass this class and
    implement `parse` and `serialize` class methods.

    ScalarMeta metaclass sets `__type_name__` attribute during class creation.
    """

    @classmethod
    def parse(cls, value: Any) -> Any:
        return value

    @classmethod
    def serialize(cls, value: Any) -> Any:
        return value

    @classmethod
    def accept(cls, visitor: "AbstractGraphVisitor") -> Any:
        return visitor.visit_scalar(cls)


class DateTime(Scalar):
    """
    The `DateTime` scalar type represents a datetime.datetime as UTC string.
    """

    @classmethod
    def parse(cls, value: str) -> datetime:
        return datetime.fromisoformat(value)

    @classmethod
    def serialize(cls, value: datetime) -> str:
        return value.isoformat()


class Date(Scalar):
    @classmethod
    def parse(cls, value: str) -> date:
        return date.fromisoformat(value)

    @classmethod
    def serialize(cls, value: date) -> str:
        return value.isoformat()


@scalar(description="The `UUID` scalar type represents a UUID. ")
class UUID(Scalar):
    @classmethod
    def parse(cls, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    @classmethod
    def serialize(cls, value: uuid.UUID) -> str:
        return str(value)



from typing import Any, Callable


class cached_property:
    def __init__(self, func: Callable) -> None:
        self.__doc__ = getattr(func, "__doc__")
        self.func = func

    def __get__(self, obj: Any, cls: Any) -> Any:
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, List, Optional, TypeVar
from typing_extensions import dataclass_transform

from hiku.directives import (
    SchemaDirectiveInfo,
    Location,
    SchemaDirective,
    SchemaDirectiveField,
    schema_directive_field,
    get_fields,
    wrap_dataclass,
)
from hiku.federation.scalars import FieldSet, LinkImport
from hiku.utils.typing import builtin_to_introspection_type

T = TypeVar("T", bound="FederationSchemaDirective")


LinkPurpose = Enum("link__Purpose", ["SECURITY", "EXECUTION"])


@dataclass
class ComposeOptions:
    import_url: Optional[str]


@dataclass
class FederationSchemaDirectiveInfo(SchemaDirectiveInfo):
    compose_options: Optional[ComposeOptions] = None


@dataclass_transform(field_specifiers=(SchemaDirectiveField,))
def schema_directive(
    *,
    name: str,
    locations: List[Location],
    description: Optional[str] = None,
    repeatable: bool = False,
    compose: bool = False,
    import_url: Optional[str] = None,
) -> Callable[..., T]:
    """Decorator to mark class as federated schema directive.
    A class with a @schema_directive decorator will become a dataclass.

    :param compose: If True, the directive will be added to the schema
        and will be wrapped in @composeDirective
        See: https://www.apollographql.com/docs/federation/federated-types/federated-directives#composedirective  # noqa: E501
    :param import_url: The URL of the subgraph that defines the directive.
        See: https://specs.apollo.dev/link/v1.0/#@link.url
    """

    def _wrap(cls: T) -> T:
        if not issubclass(cls, FederationSchemaDirective):  # type: ignore[arg-type]  # noqa: E501
            raise TypeError(
                f"{cls} is not a subclass of {FederationSchemaDirective}"
            )

        cls = wrap_dataclass(cls)  # type: ignore[arg-type, assignment]
        fields = get_fields(cls, SchemaDirectiveField)  # type: ignore[arg-type]
        for field in fields:
            field.type_ident = builtin_to_introspection_type(field.type)
            field.field_name = field.field_name or field.name

        cls.__directive_info__ = FederationSchemaDirectiveInfo(
            name=name,
            locations=locations,
            args=fields,
            description=description,
            repeatable=repeatable,
            compose_options=ComposeOptions(import_url=import_url)
            if compose
            else None,
        )

        return cls

    return _wrap


class FederationSchemaDirective(SchemaDirective):
    """Base class for federation directives"""

    __directive_info__: Any  # FederationSchemaDirectiveInfo


# v1/v2 directives
@schema_directive(
    name="key",
    locations=[Location.OBJECT, Location.INTERFACE],
    description=(
        "The @key directive is used to indicate a combination "
        "of fields that can be used to uniquely identify and "
        "fetch an object or interface."
    ),
    repeatable=True,
)
class Key(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#key
    """

    fields: FieldSet = schema_directive_field(
        description=(
            "A combination of fields that can be used to uniquely "
            "identify and fetch an object or interface"
        ),
    )
    resolvable: bool = schema_directive_field(
        description="",
        default_value=True,
    )


@schema_directive(
    name="provides",
    locations=[Location.FIELD_DEFINITION],
    description=(
        "The @provides directive is used to annotate the expected returned "
        "fieldset from a field on a base type that is guaranteed to be "
        "selectable by the gateway"
    ),
)
class Provides(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#provides
    """

    fields: FieldSet = schema_directive_field(
        description=(
            "Expected returned fieldset from a field on a base type "
            "that is guaranteed to be selectable by the gateway"
        ),
    )


@schema_directive(
    name="requires",
    locations=[Location.FIELD_DEFINITION],
    description=(
        "The @requires directive is used to annotate the required input "
        "fieldset from a base type for a resolver."
    ),
)
class Requires(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#requires
    """

    fields: FieldSet = schema_directive_field(
        description=(
            "The required input fieldset from a base type for a " "resolver"
        ),
    )


@schema_directive(
    name="external",
    locations=[Location.FIELD_DEFINITION, Location.OBJECT],
    description=(
        "The @external directive is used to mark a field "
        "as owned by another service."
    ),
)
class External(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federation-spec/#external
    """


@schema_directive(
    name="extends",
    locations=[Location.OBJECT, Location.INTERFACE],
    description=(
        "Indicates that an object or interface definition "
        "is an extension of another definition of that same type. "
    ),
)
class Extends(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#extends

    NOTE: This directive is not exposed in introspection and sdl, and used to
    specify that a type is an extension of another type.
    Apollo recommends to use `extend type` instead of @extends directive.
    """


# v2 directives
@schema_directive(
    name="tag",
    locations=[
        Location.FIELD_DEFINITION,
        Location.INTERFACE,
        Location.OBJECT,
        Location.UNION,
        Location.ARGUMENT_DEFINITION,
        Location.SCALAR,
        Location.ENUM,
        Location.ENUM_VALUE,
        Location.INPUT_OBJECT,
        Location.INPUT_FIELD_DEFINITION,
    ],
    description="Applies arbitrary string metadata to a schema location",
    repeatable=True,
)
class Tag(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#tag
    """

    name: str = schema_directive_field(
        description="The tag name to apply",
    )


@schema_directive(
    name="override",
    locations=[Location.FIELD_DEFINITION],
    description=(
        "Indicates that an object or interface definition "
        "is an extension of another definition of that same type. "
    ),
)
class Override(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#override
    """

    from_: str = schema_directive_field(
        "from",
        description=(
            "The name of the other subgraph that no longer "
            "resolves the field"
        ),
    )


@schema_directive(
    name="inaccessible",
    locations=[
        Location.FIELD_DEFINITION,
        Location.INTERFACE,
        Location.OBJECT,
        Location.UNION,
        Location.ARGUMENT_DEFINITION,
        Location.SCALAR,
        Location.ENUM,
        Location.ENUM_VALUE,
        Location.INPUT_OBJECT,
        Location.INPUT_FIELD_DEFINITION,
    ],
    description=(
        "Indicates that a definition in the subgraph schema "
        "should be omitted from the router's API schema, "
        "even if that definition is also present in other subgraphs"
    ),
)
class Inaccessible(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#inaccessible
    """


@schema_directive(
    name="shareable",
    locations=[
        Location.FIELD_DEFINITION,
        Location.OBJECT,
    ],
    description=(
        "Indicates that an object type's field is allowed "
        "to be resolved by multiple subgraphs"
    ),
)
class Shareable(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#shareable
    """


@schema_directive(
    name="interfaceObject",
    locations=[
        Location.OBJECT,
    ],
    description=(
        "Indicates that an object definition serves "
        "as an abstraction of another subgraph's entity interface. "
    ),
)
class InterfaceObject(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#interfaceobject
    """


@schema_directive(
    name="composeDirective",
    locations=[
        Location.SCHEMA,
    ],
    description=(
        "Indicates to composition that all uses of a "
        "particular custom type system directive in the "
        "subgraph schema should be preserved in the supergraph schema"
    ),
    repeatable=True,
)
class ComposeDirective(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#composedirective
    """

    name: str = schema_directive_field(
        description="The name (including the leading @) of the "
        "directive to preserve during composition.",
    )


@schema_directive(
    name="link",
    locations=[Location.SCHEMA],
    description=(
        "Indicates to composition that all uses of a "
        "particular custom type system directive in the "
        "subgraph schema should be preserved in the supergraph schema"
    ),
    repeatable=True,
)
class Link(FederationSchemaDirective):
    """
    https://www.apollographql.com/docs/federation/federated-types/federated-directives#the-link-directive
    """

    url: str = schema_directive_field()
    as_: Optional[str] = schema_directive_field(
        "as",
        default_value=None,
    )
    for_: Optional[LinkPurpose] = schema_directive_field(
        "for",
        default_value=None,
    )
    import_: Optional[List[Optional[LinkImport]]] = schema_directive_field(
        "import",
        default_value=None,
    )



"""
    hiku.graph
    ~~~~~~~~~~

    Graphs are defined by nodes, fields and links. Simple functions
    are used to fetch any data from any data source.

"""
import dataclasses
import typing as t

from abc import ABC, abstractmethod
from enum import Enum
from itertools import chain
from functools import reduce
from collections import OrderedDict, defaultdict
from typing import List

from hiku.enum import BaseEnum
from .scalar import Scalar, ScalarMeta

from .types import (
    EnumRefMeta,
    InterfaceRef,
    InterfaceRefMeta,
    Optional,
    OptionalMeta,
    RefMeta,
    Sequence,
    SequenceMeta,
    TypeRef,
    Record,
    Any,
    GenericMeta,
    TypeRefMeta,
    TypingMeta,
    AnyMeta,
    UnionRef,
    UnionRefMeta,
)
from .utils import (
    cached_property,
    const,
    Const,
)
from .directives import Deprecated, SchemaDirective

from .compat import TypeAlias


if t.TYPE_CHECKING:
    from .sources.graph import SubGraph
    from .sources.graph import BoundExpr

# TODO enum ???
Maybe = const("Maybe")

One = const("One")

Many = const("Many")

MaybeMany = const("MaybeMany")

#: Special constant that is used by links with :py:class:`~hiku.types.Optional`
#: type in order to indicate that there is nothing to link to
Nothing = const("Nothing")
NothingType: TypeAlias = Const


class AbstractBase(ABC):
    @abstractmethod
    def accept(self, visitor: "AbstractGraphVisitor") -> None:
        pass


class AbstractOption(AbstractBase, ABC):
    pass


class Option(AbstractOption):
    """Defines an option, used to customize results of the fields and links

    Options without default value are **required**.

    Example of a required option::

        Option('id', Integer)

    Example of an optional option::

        Option('size', Integer, default=100)

    Example with TypeRef type(ref can point either to Node or data type)::

        Option('filter', TypeRef['FilterInput'])

    """

    def __init__(
        self,
        name: str,
        type_: t.Optional[GenericMeta],
        *,
        default: t.Any = Nothing,
        description: t.Optional[str] = None,
    ):
        """
        :param name: name of the option
        :param type_: type of the option or ``None``
        :param default: default option value
        :param description: description of the option
        """
        self.name = name
        self.type = type_
        self.default = default
        self.description = description
        self.type_info = get_field_info(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.type
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_option(self)


class AbstractField(AbstractBase, ABC):
    pass


R = t.TypeVar("R")

SyncAsync = t.Union[R, t.Awaitable[R]]

# (fields) -> []
RootFieldFunc = t.Callable[[t.List["Field"]], SyncAsync[List[t.Any]]]
# (fields, ids) -> [[]]
NotRootFieldFunc = t.Callable[
    [t.List["Field"], List[t.Any]], SyncAsync[List[List[t.Any]]]
]
# (ctx, fields, ids) -> [[]]
NotRootFieldFuncCtx = t.Callable[
    [t.Any, t.List["Field"], List[t.Any]], SyncAsync[List[List[t.Any]]]
]


FieldT = t.Optional[t.Union[GenericMeta, ScalarMeta]]
FieldFunc = t.Union[
    RootFieldFunc,
    NotRootFieldFunc,
    NotRootFieldFuncCtx,
    "SubGraph",
    "BoundExpr",
]


class Field(AbstractField):
    """Defines a field of the node

    Example::

        graph = Graph([
            Node('user', [
                Field('name', String, func),
            ]),
        ])

    Example with options::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)],
                      directives=[Deprecated('use another field')]),
            ]),
        ])

    Data loading protocol::

        # root node fields
        def func(fields) -> List[T]

        # non-root node fields
        def func(fields, ids) -> List[List[T]]

    Where:

    - ``fields`` - list of :py:class:`hiku.query.Field`
    - ``ids`` - list node identifiers

    """

    def __init__(
        self,
        name: str,
        type_: FieldT,
        func: FieldFunc,
        *,
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        """
        :param str name: name of the field
        :param type_: type of the field or ``None``
        :param func: function to load field's data
        :param options: list of acceptable options
        :param description: description of the field
        :param directives: list of directives for the field
        :param deprecated: deprecation reason
        """
        if directives is None:
            directives = []

        if deprecated is not None:
            directives.append(Deprecated(deprecated))

        self.name = name
        self.type = type_
        self.func = func
        self.options = options or ()
        self.description = description
        self.directives = directives
        self.type_info = get_field_info(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, {!r})".format(
            self.__class__.__name__, self.name, self.type, self.func
        )

    @cached_property
    def options_map(self) -> OrderedDict:
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_field(self)


class AbstractLink(AbstractBase, ABC):
    pass


def get_link_type_enum(type_: TypingMeta) -> t.Tuple[Const, str]:
    if isinstance(type_, RefMeta):
        return One, type_.__type_name__
    elif isinstance(type_, OptionalMeta):
        if isinstance(type_.__type__, RefMeta):
            return Maybe, type_.__type__.__type_name__
    elif isinstance(type_, SequenceMeta):
        if isinstance(type_.__item_type__, RefMeta):
            return Many, type_.__item_type__.__type_name__
        elif isinstance(type_.__item_type__, OptionalMeta):
            if isinstance(type_.__item_type__.__type__, RefMeta):
                return MaybeMany, type_.__item_type__.__type__.__type_name__
    raise TypeError("Invalid type specified: {!r}".format(type_))


class FieldType(Enum):
    # When Field is a TypeRef to Record
    RECORD = "RECORD"
    # When Field is a EnumRef to Enum
    ENUM = "ENUM"
    # When Field is a builtin scalar (such as String, Integer, etc)
    SCALAR = "SCALAR"
    # When Field is a custom scalar (such as DateTime, etc)
    CUSTOM_SCALAR = "CUSTOM_SCALAR"


@dataclasses.dataclass
class FieldTypeInfo:
    type_name: str
    type_enum: FieldType


def get_field_info(
    type_: t.Optional[t.Union[GenericMeta, ScalarMeta]]
) -> t.Optional[FieldTypeInfo]:
    if not type_:
        return None

    if isinstance(type_, OptionalMeta):
        return get_field_info(getattr(type_, "__type__", None))

    if isinstance(type_, SequenceMeta):
        # if type not passed to Sequence[]
        return get_field_info(getattr(type_, "__item_type__", None))

    if isinstance(type_, TypeRefMeta):
        return FieldTypeInfo(
            type_.__type_name__,
            FieldType.RECORD,
        )
    elif isinstance(type_, EnumRefMeta):
        return FieldTypeInfo(type_.__type_name__, FieldType.ENUM)
    elif isinstance(type_, ScalarMeta):
        return FieldTypeInfo(type_.__type_name__, FieldType.CUSTOM_SCALAR)

    if isinstance(type_, GenericMeta):
        return FieldTypeInfo(type_.__name__, FieldType.SCALAR)

    return None


class LinkType(Enum):
    NODE = "NODE"
    INTERFACE = "INTERFACE"
    UNION = "UNION"


@dataclasses.dataclass
class LinkTypeInfo:
    type_name: str
    type_enum: LinkType


def get_link_type(type_: GenericMeta) -> LinkTypeInfo:
    if isinstance(type_, OptionalMeta):
        return get_link_type(type_.__type__)

    if isinstance(type_, SequenceMeta):
        return get_link_type(type_.__item_type__)

    if isinstance(type_, TypeRefMeta):
        return LinkTypeInfo(
            type_.__type_name__,
            LinkType.NODE,
        )
    elif isinstance(type_, UnionRefMeta):
        return LinkTypeInfo(type_.__type_name__, LinkType.UNION)
    elif isinstance(type_, InterfaceRefMeta):
        return LinkTypeInfo(type_.__type_name__, LinkType.INTERFACE)

    raise TypeError("Invalid link type specified: {!r}".format(type_))


def collect_interfaces_types(
    items: t.List["Node"], interfaces: t.List["Interface"]
) -> t.Dict[str, t.List[str]]:
    interfaces_types = defaultdict(list)
    for item in items:
        if item.name is not None and item.implements:
            for impl in item.implements:
                interfaces_types[impl].append(item.name)

    for i in interfaces:
        if i.name not in interfaces_types:
            interfaces_types[i.name] = []

    return dict(interfaces_types)


LT = t.TypeVar("LT", bound=t.Hashable)
LR = t.TypeVar("LR", bound=t.Optional[t.Hashable])

MaybeLink = t.Union[LR, NothingType]
RootLinkT = t.Union[
    # () -> LR
    t.Callable[[], SyncAsync[LR]],
    # (opts) -> LR
    t.Callable[[t.Any], SyncAsync[LR]],
    # (ctx, opts) -> LR
    t.Callable[[t.Any, t.Any], SyncAsync[LR]],
]

LinkT = t.Union[
    # (ids) -> []
    t.Callable[[List[LT]], SyncAsync[LR]],
    # (ids, opts) -> []
    t.Callable[[List[LT], t.Any], SyncAsync[LR]],
    # (ctx, ids) -> []
    t.Callable[[t.Any, List[LT]], SyncAsync[LR]],
    # (ctx, ids, opts) -> []
    t.Callable[[t.Any, List[LT], t.Dict], SyncAsync[LR]],
]

RootLinkOne = RootLinkT[LR]
RootLinkMaybe = RootLinkT[MaybeLink[LR]]
RootLinkMany = RootLinkT[List[LR]]  # type: ignore[type-var]

LinkOne = LinkT[LT, List[LR]]  # type: ignore[type-var]
LinkMaybe = LinkT[LT, List[MaybeLink[LR]]]  # type: ignore[type-var]
LinkMany = LinkT[LT, List[List[LR]]]  # type: ignore[type-var]

LinkFunc = t.Union[
    RootLinkOne,
    RootLinkMaybe,
    RootLinkMany,
    LinkOne,
    LinkMaybe,
    LinkMany,
]

LinkOneFunc = t.Union[RootLinkOne, LinkOne]
LinkMaybeFunc = t.Union[RootLinkMaybe, LinkMaybe]
LinkManyFunc = t.Union[RootLinkMany, LinkMany]


class Link(AbstractLink):
    """Defines a link to the node

    Example::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None),
            ]),
        ])

    Example with requirements::

        graph = Graph([
            Node('character', [...]),
            Node('actor', [
                Field('id', Integer, field_func),
                Link('characters', Sequence[TypeRef['character']],
                     link_func, requires='id'),
            ])
        ])

    Requirements are needed when link points from non-root node.

    Example with options::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)],
                     directives=[Deprecated('do not use')]),
            ]),
        ])

    Identifiers loading function protocol::

        # From root node or if requires is None:

        # ... if type is TypeRef['foo']
        def func() -> T

        # ... if type is Optional[TypeRef['foo']]
        def func() -> Union[T, Nothing]

        # ... if type is Sequence[TypeRef['foo']]
        def func() -> List[T]

        # From non-root node and requires is not None:

        # ... if type is TypeRef['foo']
        def func(ids) -> List[T]

        # ... if type is Optional[TypeRef['foo']]
        def func(ids) -> List[Union[T, Nothing]]

        # ... if type is Sequence[TypeRef['foo']]
        def func(ids) -> List[List[T]]

    See also :py:const:`hiku.graph.Nothing`.

    If link was defined with options, then link function should accept one
    additional positional argument::

        # many to many link with options
        def func(ids, options) -> List[List[T]]

    Where ``options`` is a mapping ``str: value`` of provided in the query
    options.
    """

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[TypeRef],
        func: LinkOneFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[t.Union[UnionRef, InterfaceRef]],
        func: LinkOneFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[Optional],
        func: LinkMaybeFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[Sequence],
        func: LinkManyFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        ...

    def __init__(  # type: ignore[no-untyped-def]
        self,
        name,
        type_,
        func,
        *,
        requires,
        options=None,
        description=None,
        directives=None,
        deprecated=None,
    ):
        """
        :param name: name of the link
        :param type_: type of the link
        :param func: function to load identifiers of the linked node
        :param requires: field name(s) from the current node,
            required to compute identifiers of the linked node
        :param options: list of acceptable options
        :param description: description of the link
        :param directives: list of directives for the link
        :param deprecated: deprecation reason for the link
        """
        type_enum, node = get_link_type_enum(type_)

        if directives is None:
            directives = []

        if deprecated is not None:
            directives.append(Deprecated(deprecated))

        self.name = name
        self.type = type_
        self.type_enum = type_enum
        self.node = node
        self.func = func
        self.requires = requires
        self.options = options or ()
        self.description = description
        self.directives = directives
        self.type_info = get_link_type(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.type, self.func
        )

    @cached_property
    def options_map(self) -> OrderedDict:
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_link(self)


class AbstractNode(AbstractBase, ABC):
    pass


class Union(AbstractBase):
    def __init__(
        self,
        name: str,
        types: t.List[str],
        *,
        description: t.Optional[str] = None,
    ):
        self.name = name
        self.types = types
        self.description = description

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.types
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_union(self)


class Interface(AbstractBase):
    def __init__(
        self,
        name: str,
        fields: t.Sequence[t.Union["Field", "Link"]],
        *,
        description: t.Optional[str] = None,
    ):
        self.name = name
        self.fields = fields
        self.description = description

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.fields
        )

    @cached_property
    def fields_map(self) -> OrderedDict:
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_interface(self)


class Node(AbstractNode):
    """Collection of the fields and links, which describes some entity and
    relations with other entities

    Example::

        graph = Graph([
            Node('user', [
                Field('id', Integer, field_func),
                Field('name', String, field_func),
                Link('roles', Sequence[TypeRef['role']],
                     link_func, requires='id'),
            ]),
        ])

    """

    def __init__(
        self,
        name: t.Optional[str],
        fields: t.List[t.Union[Field, Link]],
        *,
        description: t.Optional[str] = None,
        directives: t.Optional[t.Sequence[SchemaDirective]] = None,
        implements: t.Optional[t.Sequence[str]] = None,
    ):
        """
        :param name: name of the node
        :param fields: list of fields and links
        :param description: description of the node
        :param directives: list of directives for the node
        :param implements: list of interfaces implemented by the node
        """
        self.name = name
        self.fields = fields
        self.description = description
        self.directives: t.Tuple[SchemaDirective, ...] = tuple(directives or ())
        self.implements = tuple(implements or [])

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.fields
        )

    @cached_property
    def fields_map(self) -> OrderedDict:
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_node(self)

    def copy(self) -> "Node":
        return Node(
            name=self.name,
            fields=self.fields[:],
            description=self.description,
            directives=self.directives,
            implements=self.implements,
        )


class Root(Node):
    """Special implicit root node, starting point of the query execution

    Example::

        graph = Graph([
            Node('baz', [...]),
            Root([
                Field('foo', String, root_fields_func),
                Link('bar', Sequence[TypeRef['baz']],
                     to_baz_func, requires=None),
            ]),
        ])

    """

    def __init__(
        self,
        items: t.List[t.Union[Field, Link]],
    ):
        """
        :param items: list of fields, links and singleton nodes
        """
        super(Root, self).__init__(None, items)

    def __repr__(self) -> str:
        return "{}({!r}, {!r})".format(
            self.__class__.__name__, self.name, self.fields
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_root(self)


class AbstractGraph(AbstractBase, ABC):
    pass


G = t.TypeVar("G", bound="Graph")


class Graph(AbstractGraph):
    """Collection of nodes - definition of the graph

    Example::

        graph = Graph([
            Node('foo', [...]),
            Node('bar', [...]),
            Root([...]),
        ])

    """

    def __init__(
        self,
        items: t.List[Node],
        data_types: t.Optional[t.Dict[str, t.Type[Record]]] = None,
        directives: t.Optional[t.Sequence[t.Type[SchemaDirective]]] = None,
        unions: t.Optional[t.List[Union]] = None,
        interfaces: t.Optional[t.List[Interface]] = None,
        enums: t.Optional[t.List[BaseEnum]] = None,
        scalars: t.Optional[t.List[t.Type[Scalar]]] = None,
    ):
        """
        :param items: list of nodes
        """
        from .validate.graph import GraphValidator

        if unions is None:
            unions = []

        if interfaces is None:
            interfaces = []

        if enums is None:
            enums = []

        if scalars is None:
            scalars = []

        GraphValidator.validate(items, unions, interfaces, enums, scalars)

        self.items = GraphInit.init(items)
        self.unions = unions
        self.interfaces = interfaces
        self.interfaces_types = collect_interfaces_types(self.items, interfaces)
        self.enums: t.List[BaseEnum] = enums
        self.scalars = scalars
        self.data_types = data_types or {}
        self.__types__ = GraphTypes.get_types(
            self.items,
            self.unions,
            self.interfaces,
            self.enums,
            self.data_types,
        )
        self.directives: t.Tuple[t.Type[SchemaDirective], ...] = tuple(
            directives or ()
        )

    def __repr__(self) -> str:
        return "{}({!r})".format(self.__class__.__name__, self.items)

    def iter_root(self) -> t.Iterator[t.Union[Field, Link]]:
        for node in self.items:
            if node.name is None:
                for field in node.fields:
                    yield field

    def iter_nodes(self) -> t.Iterator[Node]:
        for node in self.items:
            if node.name is not None:
                yield node

    @cached_property
    def root(self) -> Root:
        return Root(list(self.iter_root()))

    @cached_property
    def nodes(self) -> t.List[Node]:
        return list(self.iter_nodes())

    @cached_property
    def nodes_map(self) -> OrderedDict:
        return OrderedDict((e.name, e) for e in self.iter_nodes())

    @cached_property
    def unions_map(self) -> OrderedDict:
        return OrderedDict((u.name, u) for u in self.unions)

    @cached_property
    def interfaces_map(self) -> OrderedDict:
        return OrderedDict((i.name, i) for i in self.interfaces)

    @cached_property
    def enums_map(self) -> "OrderedDict[str, BaseEnum]":
        return OrderedDict((e.name, e) for e in self.enums)

    @cached_property
    def scalars_map(self) -> "OrderedDict[str, t.Type[Scalar]]":
        return OrderedDict((s.__type_name__, s) for s in self.scalars)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_graph(self)

    @classmethod
    def from_graph(cls: t.Type[G], other: G, root: Root) -> G:
        """Create graph from other graph, with new root node.
        Useful for creating mutation graph from query graph.
        Example:
            MUTATION_GRAPH = Graph.from_graph(QUERY_GRAPH, Root([...]))
        """
        return cls(
            other.nodes + [root],
            data_types=other.data_types,
            directives=other.directives,
            unions=other.unions,
            interfaces=other.interfaces,
            enums=other.enums,
            scalars=other.scalars,
        )


class AbstractGraphVisitor(ABC):
    @abstractmethod
    def visit(self, obj: t.Any) -> t.Any:
        pass

    @abstractmethod
    def visit_option(self, obj: Option) -> t.Any:
        pass

    @abstractmethod
    def visit_field(self, obj: Field) -> t.Any:
        pass

    @abstractmethod
    def visit_link(self, obj: Link) -> t.Any:
        pass

    @abstractmethod
    def visit_node(self, obj: Node) -> t.Any:
        pass

    @abstractmethod
    def visit_union(self, obj: Union) -> t.Any:
        pass

    @abstractmethod
    def visit_interface(self, obj: Interface) -> t.Any:
        pass

    @abstractmethod
    def visit_enum(self, obj: BaseEnum) -> t.Any:
        pass

    @abstractmethod
    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Any:
        pass

    @abstractmethod
    def visit_root(self, obj: Root) -> t.Any:
        pass

    @abstractmethod
    def visit_graph(self, obj: Graph) -> t.Any:
        pass


class GraphVisitor(AbstractGraphVisitor):
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_union(self, obj: "Union") -> t.Any:
        pass

    def visit_interface(self, obj: "Interface") -> t.Any:
        pass

    def visit_enum(self, obj: "BaseEnum") -> t.Any:
        pass

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Any:
        pass

    def visit_option(self, obj: "Option") -> t.Any:
        pass

    def visit_field(self, obj: Field) -> t.Any:
        for option in obj.options:
            self.visit(option)

    def visit_link(self, obj: Link) -> t.Any:
        for option in obj.options:
            self.visit(option)

    def visit_node(self, obj: Node) -> t.Any:
        for item in obj.fields:
            self.visit(item)

    def visit_root(self, obj: Root) -> t.Any:
        for item in obj.fields:
            self.visit(item)

    def visit_graph(self, obj: Graph) -> t.Any:
        for item in obj.items:
            self.visit(item)


class GraphTransformer(AbstractGraphVisitor):
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_option(self, obj: Option) -> Option:
        return Option(
            obj.name, obj.type, default=obj.default, description=obj.description
        )

    def visit_field(self, obj: Field) -> Field:
        return Field(
            obj.name,
            obj.type,
            obj.func,
            options=[self.visit(op) for op in obj.options],
            description=obj.description,
            directives=obj.directives,
        )

    def visit_link(self, obj: Link) -> Link:
        return Link(
            obj.name,
            obj.type,
            obj.func,
            requires=obj.requires,
            options=[self.visit(op) for op in obj.options],
            description=obj.description,
            directives=obj.directives,
        )

    def visit_node(self, obj: Node) -> Node:
        return Node(
            obj.name,
            [self.visit(f) for f in obj.fields],
            description=obj.description,
            directives=obj.directives,
            implements=obj.implements,
        )

    def visit_union(self, obj: Union) -> Union:
        return Union(
            obj.name,
            obj.types,
            description=obj.description,
        )

    def visit_interface(self, obj: Interface) -> Interface:
        return Interface(
            obj.name,
            obj.fields,
            description=obj.description,
        )

    def visit_enum(self, obj: BaseEnum) -> BaseEnum:
        return obj

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Type[Scalar]:
        return obj

    def visit_root(self, obj: Root) -> Root:
        return Root([self.visit(f) for f in obj.fields])

    def visit_graph(self, obj: Graph) -> Graph:
        return Graph(
            [self.visit(node) for node in obj.items],
            obj.data_types,
            obj.directives,
            obj.unions,
            obj.interfaces,
            obj.enums,
            obj.scalars,
        )


def apply(graph: G, transformers: List[GraphTransformer]) -> G:
    """Helper function to apply graph transformations

    Example:

    .. code-block:: python

        graph = hiku.graph.apply(graph, [AsyncGraphQLIntrospection()])

    """
    return reduce(lambda g, tr: tr.visit(g), transformers, graph)


class GraphInit(GraphTransformer):
    @classmethod
    def init(cls, items: t.List[Node]) -> t.List[Node]:
        self = cls()
        return [self.visit(i) for i in items]

    def visit_field(self, obj: Field) -> Field:
        field = super(GraphInit, self).visit_field(obj)
        postprocess = getattr(field.func, "__postprocess__", None)
        if postprocess is not None:
            postprocess(field)
        return field

    def visit_link(self, obj: Link) -> Link:
        link = super(GraphInit, self).visit_link(obj)
        postprocess = getattr(link.func, "__postprocess__", None)
        if postprocess is not None:
            postprocess(link)
        return link


class GraphTypes(GraphVisitor):
    def _visit_graph(
        self,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        data_types: t.Dict[str, t.Type[Record]],
    ) -> t.Dict[str, t.Type[Record]]:
        types = OrderedDict(data_types)
        roots = []
        for item in items:
            if item.name is not None:
                types[item.name] = self.visit(item)
            else:
                roots.append(self.visit(item))

        for union in unions:
            types[union.name] = self.visit(union)

        for interface in interfaces:
            types[interface.name] = self.visit(interface)

        for enum in enums:
            types[enum.name] = self.visit(enum)

        types["__root__"] = Record[
            chain.from_iterable(r.__field_types__.items() for r in roots)
        ]
        return types

    @classmethod
    def get_types(
        cls,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        data_types: t.Dict[str, t.Type[Record]],
    ) -> t.Dict[str, t.Type[Record]]:
        return cls()._visit_graph(items, unions, interfaces, enums, data_types)

    def visit_graph(self, obj: Graph) -> t.Dict[str, t.Type[Record]]:
        return self._visit_graph(
            obj.items, obj.unions, obj.interfaces, obj.enums, obj.data_types
        )

    def visit_node(self, obj: Node) -> t.Type[Record]:
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_root(self, obj: Root) -> t.Type[Record]:
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_link(
        self, obj: Link
    ) -> t.Union[t.Type[TypeRef], t.Type[Optional], t.Type[Sequence]]:
        return obj.type

    def visit_field(self, obj: Field) -> t.Union[FieldT, AnyMeta]:
        return obj.type or Any

    def visit_union(self, obj: Union) -> Union:
        return obj

    def visit_enum(self, obj: BaseEnum) -> BaseEnum:
        return obj

    def visit_interface(self, obj: Interface) -> Interface:
        return obj

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Type[Scalar]:
        return obj



from typing import (
    TypeVar,
    Callable,
    TYPE_CHECKING,
    Any,
)

from hiku.executors.base import BaseSyncExecutor
from hiku.result import Proxy

if TYPE_CHECKING:
    from hiku.executors.queue import (
        Queue,
        Workflow,
    )


T = TypeVar("T")


class FutureLike:
    def __init__(self, result: T) -> None:
        self._result = result

    def result(self) -> T:  # type: ignore[type-var]
        return self._result


class SyncExecutor(BaseSyncExecutor):
    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike:
        return FutureLike(fn(*args, **kwargs))

    def process(self, queue: "Queue", workflow: "Workflow") -> Proxy:
        while queue.__futures__:
            queue.progress(queue.__futures__)
        return workflow.result()



import sys

from enum import EnumMeta
from typing import Any, Type, Union

from hiku.introspection import types as int_types
from hiku.scalar import Scalar


def is_union(annotation: object) -> bool:
    # this check is needed because unions declared with the new syntax `A | B`
    # don't have a `__origin__` property on them, but they are instances of
    # `UnionType`, which is only available in Python 3.10+
    if sys.version_info >= (3, 10):
        from types import UnionType

        if isinstance(annotation, UnionType):
            return True

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == Union


def is_optional(annotation: Type) -> bool:
    """Returns True if the annotation is Optional[SomeType]"""

    # Optionals are represented as unions

    if not is_union(annotation):
        return False

    types_ = annotation.__args__

    # A Union to be optional needs to have at least one None type
    return any(x == None.__class__ for x in types_)  # noqa: E711


def is_list(annotation: object) -> bool:
    """Returns True if annotation is a List"""

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == list


def builtin_to_introspection_type(typ: Any) -> Any:
    def convert(typ_: Any) -> Any:  # SCALAR
        if typ_ is int:
            return int_types.SCALAR("Int")
        elif typ_ is str:
            return int_types.SCALAR("String")
        elif typ_ is bool:
            return int_types.SCALAR("Boolean")
        elif typ_ is float:
            return int_types.SCALAR("Float")
        elif is_optional(typ_):
            return convert(typ_.__args__[0])
        elif is_list(typ_):
            return int_types.LIST(convert(typ_.__args__[0]))
        elif issubclass(typ_, Scalar):
            return int_types.SCALAR(typ_.__type_name__)
        elif isinstance(typ_, EnumMeta):
            return int_types.ENUM(
                typ_.__name__,
                [v for v in typ_.__members__],
            )
        else:
            raise TypeError(f"Unknown type: {typ_}")

    optional = False
    if is_optional(typ):
        optional = True

    intr_type = convert(typ)

    if optional:
        return intr_type

    return int_types.NON_NULL(intr_type)



from functools import partial
from collections import defaultdict
from typing import (
    Optional,
    Union,
    List,
    Any,
    Tuple,
    Callable,
    Dict,
    Iterable,
)

import sqlalchemy
from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import BinaryExpression

from ..types import (
    String,
    Integer,
    IntegerMeta,
    StringMeta,
)
from ..graph import (
    Nothing,
    Maybe,
    One,
    Many,
    Link,
    Field,
)
from ..query import Field as QueryField
from ..engine import (
    pass_context,
    Context,
)

SQLALCHEMY_VERSION = tuple(map(int, sqlalchemy.__version__.split(".")))

if SQLALCHEMY_VERSION >= (1, 4):
    from sqlalchemy.engine.row import Row
else:
    from sqlalchemy.engine import RowProxy as Row


def _translate_type(
    column: sqlalchemy.Column,
) -> Optional[Union[IntegerMeta, StringMeta]]:
    if isinstance(column.type, sqlalchemy.Integer):
        return Integer
    elif isinstance(column.type, sqlalchemy.Unicode):
        return String
    else:
        return None


def _table_repr(table: sqlalchemy.Table) -> str:
    return "Table({})".format(
        ", ".join(
            [
                repr(table.name),
                repr(table.metadata),
                "...",
                "schema={!r}".format(table.schema),
            ]
        )
    )


@pass_context
class FieldsQuery:
    def __init__(
        self,
        engine_key: str,
        from_clause: sqlalchemy.Table,
        *,
        primary_key: Optional[sqlalchemy.Column] = None,
    ) -> None:
        self.engine_key = engine_key
        self.from_clause = from_clause
        if primary_key is not None:
            self.primary_key = primary_key
        else:
            # currently only one column supported
            (self.primary_key,) = from_clause.primary_key

    def __repr__(self) -> str:
        if isinstance(self.from_clause, sqlalchemy.Table):
            from_clause_repr = _table_repr(self.from_clause)
        else:
            from_clause_repr = repr(self.from_clause)
        return (
            "<{}.{}: engine_key={!r}, from_clause={}, primary_key={!r}>".format(
                self.__class__.__module__,
                self.__class__.__name__,
                self.engine_key,
                from_clause_repr,
                self.primary_key,
            )
        )

    def __postprocess__(self, field: Field) -> None:
        if field.type is None:
            column = self.from_clause.c[field.name]
            field.type = _translate_type(column)

    def in_impl(
        self, column: sqlalchemy.Column, values: Iterable
    ) -> BinaryExpression:
        return column.in_(values)

    def select_expr(
        self, fields_: List[QueryField], ids: Iterable
    ) -> Tuple[Select, Callable]:
        columns = [self.from_clause.c[f.name] for f in fields_]
        expr = (
            sqlalchemy.select([self.primary_key] + columns)
            .select_from(self.from_clause)
            .where(self.in_impl(self.primary_key, ids))
        )

        def result_proc(rows: List[Row]) -> List:
            rows_map = {
                row[self.primary_key]: [row[c] for c in columns] for row in rows
            }

            nulls = [None for _ in fields_]
            return [rows_map.get(id_, nulls) for id_ in ids]

        return expr, result_proc

    def __call__(
        self, ctx: Context, fields_: List[QueryField], ids: List
    ) -> Any:
        if not ids:
            return []

        expr, result_proc = self.select_expr(fields_, ids)

        sa_engine = ctx[self.engine_key]
        with sa_engine.connect() as connection:
            rows = connection.execute(expr).fetchall()

        return result_proc(rows)


def _to_maybe_mapper(pairs: List[Tuple[Any, Any]], values: List) -> List:
    mapping: Dict = dict(pairs)
    return [mapping.get(value, Nothing) for value in values]


def _to_one_mapper(pairs: List[Tuple[Any, Any]], values: List) -> List:
    mapping = dict(pairs)
    return [mapping[value] for value in values]


def _to_many_mapper(pairs: List[Tuple], values: List) -> List:
    mapping = defaultdict(list)
    for from_value, to_value in pairs:
        mapping[from_value].append(to_value)
    return [mapping[value] for value in values]


class LinkQuery:
    def __init__(
        self,
        engine_key: str,
        *,
        from_column: sqlalchemy.Column,
        to_column: sqlalchemy.Column,
    ) -> None:
        if from_column.table is not to_column.table:
            raise ValueError(
                "from_column and to_column should belong to " "one table"
            )

        self.engine_key = engine_key
        self.from_column = from_column
        self.to_column = to_column

    def __repr__(self) -> str:
        return (
            "<{}.{}: engine_key={!r}, from_column={!r}, to_column={!r}>".format(
                self.__class__.__module__,
                self.__class__.__name__,
                self.engine_key,
                self.from_column,
                self.to_column,
            )
        )

    def __postprocess__(self, link: Link) -> None:
        if link.type_enum is One:
            func = partial(self, _to_one_mapper)
        elif link.type_enum is Maybe:
            func = partial(self, _to_maybe_mapper)
        elif link.type_enum is Many:
            func = partial(self, _to_many_mapper)
        else:
            raise TypeError(repr(link.type_enum))
        link.func = pass_context(func)

    def in_impl(
        self, column: sqlalchemy.Column, values: Iterable
    ) -> BinaryExpression:
        return column.in_(values)

    def select_expr(self, ids: Iterable) -> Optional[Select]:
        # TODO: make this optional, but enabled by default
        filtered_ids = [i for i in set(ids) if i is not None]
        if filtered_ids:
            return sqlalchemy.select(
                [
                    self.from_column.label("from_column"),
                    self.to_column.label("to_column"),
                ]
            ).where(self.in_impl(self.from_column, filtered_ids))
        else:
            return None

    def __call__(
        self, result_proc: Callable, ctx: Context, ids: Iterable
    ) -> Any:
        expr = self.select_expr(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.engine_key]
            with sa_engine.connect() as connection:
                pairs = connection.execute(expr).fetchall()
        return result_proc(pairs, ids)



import typing as t
from collections import deque

from ..enum import BaseEnum
from ..graph import (
    FieldType,
    Graph,
    Interface,
    Union,
    Field as GraphField,
)
from ..query import (
    QueryVisitor,
    Link,
    Field,
    Node,
)
from ..result import Proxy
from ..types import (
    Record,
    RecordMeta,
    RefMeta,
    OptionalMeta,
    SequenceMeta,
    get_type,
)
from ..utils.serialize import serialize


def serialize_value(graph: Graph, field: GraphField, value: t.Any) -> t.Any:
    """Serializes value according to the field type."""
    if not field.type_info:
        return value

    field_type = field.type_info.type_enum
    type_name = field.type_info.type_name

    if field_type in (FieldType.SCALAR, FieldType.RECORD):
        return value
    elif field_type is FieldType.ENUM:
        enum = graph.enums_map[type_name]
        return serialize(field.type, value, enum.serialize)
    elif field_type is FieldType.CUSTOM_SCALAR:
        scalar = graph.scalars_map[type_name]
        return serialize(field.type, value, scalar.serialize)
    else:
        raise TypeError(
            'Unknown field "{}" type "{!r}"'.format(field.name, field.type)
        )


class Denormalize(QueryVisitor):
    def __init__(self, graph: Graph, result: Proxy) -> None:
        self._graph = graph
        self._types = graph.__types__
        self._unions = graph.unions_map
        self._enums = graph.enums_map
        self._result = result
        self._type: t.Deque[
            t.Union[t.Type[Record], Union, Interface, BaseEnum]
        ] = deque([self._types["__root__"]])
        self._data = deque([result])
        self._res: t.Deque = deque()

    def process(self, query: Node) -> t.Dict:
        assert not self._res, self._res
        self._res.append({})
        self.visit(query)
        return self._res.pop()

    def visit_node(self, obj: Node) -> t.Any:
        for item in obj.fields:
            self.visit(item)

        type_name = None
        if isinstance(self._data[-1], Proxy):
            type_name = self._data[-1].__ref__.node

        for fr in obj.fragments:
            if type_name is not None and type_name != fr.type_name:
                # do not visit fragment if type specified and not match
                continue

            self.visit(fr)

    def visit_field(self, obj: Field) -> None:
        if isinstance(self._data[-1], Proxy):
            type_name = self._data[-1].__ref__.node
            if type_name == "__root__":
                node = self._graph.root
            else:
                node = self._graph.nodes_map[type_name]
            graph_field = node.fields_map[obj.name]

            self._res[-1][obj.result_key] = serialize_value(
                self._graph, graph_field, self._data[-1][obj.result_key]
            )
        else:
            # Record type itself does not have custom serialization
            # TODO: support Scalar/Enum types in Record
            self._res[-1][obj.result_key] = self._data[-1][obj.result_key]

    def visit_link(self, obj: Link) -> None:
        if isinstance(self._type[-1], (Union, Interface)):
            type_ = self._types[self._data[-1].__ref__.node].__field_types__[
                obj.name
            ]
        elif isinstance(self._type[-1], RecordMeta):
            type_ = self._type[-1].__field_types__[obj.name]
        else:
            raise AssertionError(repr(self._type[-1]))

        if isinstance(type_, RefMeta):
            self._type.append(get_type(self._types, type_))
            self._res.append({})
            self._data.append(self._data[-1][obj.result_key])
            super().visit_link(obj)
            self._data.pop()
            self._res[-1][obj.result_key] = self._res.pop()
            self._type.pop()
        elif isinstance(type_, SequenceMeta):
            type_ref = type_.__item_type__
            if isinstance(type_.__item_type__, OptionalMeta):
                type_ref = type_.__item_type__.__type__
            assert isinstance(type_ref, RefMeta)
            self._type.append(get_type(self._types, type_ref))
            items = []
            for item in self._data[-1][obj.result_key]:
                self._res.append({})
                self._data.append(item)
                super().visit_link(obj)
                self._data.pop()
                items.append(self._res.pop())
            self._res[-1][obj.result_key] = items
            self._type.pop()
        elif isinstance(type_, OptionalMeta):
            if self._data[-1][obj.result_key] is None:
                self._res[-1][obj.result_key] = None
            else:
                assert isinstance(type_.__type__, RefMeta)
                self._type.append(get_type(self._types, type_.__type__))
                self._res.append({})
                self._data.append(self._data[-1][obj.result_key])
                super().visit_link(obj)
                self._data.pop()
                self._res[-1][obj.result_key] = self._res.pop()
                self._type.pop()
        else:
            raise AssertionError(repr(type_))



from typing import Any, Dict, Generic, NoReturn, Optional, Set, TypeVar, cast

K = TypeVar("K")
V = TypeVar("V")


class ImmutableDict(Dict, Generic[K, V]):
    _hash = None

    def __hash__(self) -> int:  # type: ignore
        if self._hash is None:
            self._hash = hash(frozenset(self.items()))
        return self._hash

    def _immutable(self) -> NoReturn:
        raise TypeError(
            "{} object is immutable".format(self.__class__.__name__)
        )

    __delitem__ = __setitem__ = _immutable  # type: ignore
    clear = pop = popitem = setdefault = update = _immutable  # type: ignore


def to_immutable_dict(
    data: Dict[K, V],
    exclude_keys: Optional[Set[str]] = None,
) -> ImmutableDict[K, V]:
    immutable: Dict[Any, Any] = dict()
    for key in data:
        if exclude_keys and key in exclude_keys:
            continue

        if isinstance(data[key], dict):
            immutable[key] = to_immutable_dict(cast(Dict[K, V], data[key]))
        else:
            immutable[key] = data[key]

    return ImmutableDict(immutable)



from abc import abstractmethod
from typing import List

from hiku.graph import Graph
from hiku.query import Node, QueryVisitor


class QueryValidator(QueryVisitor):
    @abstractmethod
    def validate(self, query: Node, graph: Graph) -> List[str]:
        """Returns list of errors"""
        ...
